import itertools
import sys
import unittest
from textwrap import dedent

from rope.base import ast
from rope.refactor import patchedast
from ropetest import testutils

NameConstant = "Name" if sys.version_info <= (3, 8) else "NameConstant"


class PatchedASTTest(unittest.TestCase):
    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def assert_single_case_match_block(self, checker, match_type):
        checker.check_children("Match", [
            "match",
            " ",
            "Name",
            "",
            ":",
            "\n    ",
            "match_case",
        ])
        checker.check_children("match_case", [
            "case",
            " ",
            match_type,
            "",
            ":",
            "\n        ",
            "Expr",
        ])

    def test_operator_support_completeness(self):
        ast_ops = {
            n.__name__
            for n in itertools.chain(
                ast.boolop.__subclasses__(),
                ast.cmpop.__subclasses__(),
                ast.operator.__subclasses__(),
                ast.unaryop.__subclasses__(),
            )
        }
        supported_ops = set(patchedast._PatchingASTWalker._operators)
        assert ast_ops == supported_ops

    def test_bytes_string(self):
        source = '1 + b"("\n'
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        str_fragment = 'b"("'
        start = source.index(str_fragment)
        checker.check_region("Bytes", start, start + len(str_fragment))
        checker.check_children("Bytes", [str_fragment])

    def test_integer_literals_and_region(self):
        source = "a = 10\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        start = source.index("10")
        checker.check_region("Num", start, start + 2)

    def test_negative_integer_literals_and_region(self):
        source = "a = -10\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        start = source.index("-10") + 1
        end = start + 2
        # Python 3 parses as UnaryOp(op=USub(), operand=Num(n=10))
        checker.check_region("Num", start, end)

    def test_scientific_integer_literals_and_region(self):
        source = "a = -1.0e-3\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        start = source.index("-1.0e-3") + 1
        end = start + 6
        # Python 3 parses as UnaryOp(op=USub(), operand=Num(n=10))
        checker.check_region("Num", start, end)

    def test_hex_integer_literals_and_region(self):
        source = "a = 0x1\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        start = source.index("0x1")
        checker.check_region("Num", start, start + 3)

    def test_octal_integer_literals_and_region(self):
        source = "a = -0o1251\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        start = source.index("-0o1251") + 1
        end = start + 6
        # Python 3 parses as UnaryOp(op=USub(), operand=Num(n=10))
        checker.check_region("Num", start, end)
        checker.check_children("Num", ["0o1251"])

    def test_integer_literals_and_sorted_children(self):
        source = "a = 10\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Num", ["10"])

    def test_ellipsis(self):
        source = "a[...]\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        start = source.index("...")
        checker.check_region("Ellipsis", start, start + len("..."))

    def test_ass_name_node(self):
        source = "a = 10\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        start = source.index("a")
        checker.check_region("Name", start, start + 1)
        checker.check_children("Name", ["a"])

    def test_assign_node(self):
        source = "a = 10\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Assign", 0, len(source) - 1)
        checker.check_children("Assign", ["Name", " ", "=", " ", "Num"])

    @testutils.only_for_versions_higher("3.6")
    def test_ann_assign_node_without_target(self):
        source = "a: List[int]\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("AnnAssign", 0, len(source) - 1)
        checker.check_children("AnnAssign", ["Name", "", ":", " ", "Subscript"])

    @testutils.only_for_versions_higher("3.6")
    def test_ann_assign_node_with_target(self):
        source = "a: int = 10\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("AnnAssign", 0, len(source) - 1)
        checker.check_children(
            "AnnAssign", ["Name", "", ":", " ", "Name", " ", "=", " ", "Num"]
        )

    def test_add_node(self):
        source = "1 + 2\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("BinOp", 0, len(source) - 1)
        checker.check_children("BinOp", ["Num", " ", "+", " ", "Num"])

    def test_lshift_node(self):
        source = "1 << 2\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("BinOp", 0, len(source) - 1)
        checker.check_children("BinOp", ["Num", " ", "<<", " ", "Num"])

    def test_and_node(self):
        source = "True and True\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("BoolOp", 0, len(source) - 1)
        checker.check_children("BoolOp", [NameConstant, " ", "and", " ", NameConstant])

    def test_matmult_node(self):
        source = "a @ b\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("BinOp", ["Name", " ", "@", " ", "Name"])

    def test_basic_closing_parens(self):
        source = "1 + (2)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("BinOp", 0, len(source) - 1)
        checker.check_children("BinOp", ["Num", " ", "+", " (", "Num", ")"])

    def test_basic_opening_parens(self):
        source = "(1) + 2\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("BinOp", 0, len(source) - 1)
        checker.check_children("BinOp", ["(", "Num", ") ", "+", " ", "Num"])

    def test_basic_opening_biway(self):
        source = "(1) + (2)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("BinOp", 0, len(source) - 1)
        checker.check_children("BinOp", ["(", "Num", ") ", "+", " (", "Num", ")"])

    def test_basic_opening_double(self):
        source = "1 + ((2))\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("BinOp", 0, len(source) - 1)
        checker.check_children("BinOp", ["Num", " ", "+", " ((", "Num", "))"])

    def test_handling_comments(self):
        source = "(1 + #(\n2)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("BinOp", ["Num", " ", "+", " #(\n", "Num"])

    def test_handling_parens_with_spaces(self):
        source = "1 + (2\n    )\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("BinOp", ["Num", " ", "+", " (", "Num", "\n    )"])

    def test_handling_strings(self):
        source = '1 + "("\n'
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("BinOp", ["Num", " ", "+", " ", "Str"])

    def test_handling_fstrings(self):
        source = '1 + f"("\n'
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("BinOp", ["Num", " ", "+", " ", "JoinedStr"])

    def test_handling_implicit_string_concatenation(self):
        source = "a = '1''2'"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Assign", ["Name", " ", "=", " ", "Str"])
        checker.check_children("Str", ["'1''2'"])

    def test_handling_implicit_string_concatenation_line_breaks(self):
        source = dedent("""\
            a = '1' \\
            '2'""")
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Assign", ["Name", " ", "=", " ", "Str"])
        checker.check_children("Str", ["'1' \\\n'2'"])

    def test_handling_explicit_string_concatenation_line_breaks(self):
        source = "a = ('1' \n'2')"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Assign", ["Name", " ", "=", " (", "Str", ")"])
        checker.check_children("Str", ["'1' \n'2'"])

    def test_not_concatenating_strings_on_separate_lines(self):
        source = dedent("""\
            '1'
            '2'
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Module", ["", "Expr", "\n", "Expr", "\n"])

    def test_handling_raw_strings(self):
        source = 'r"abc"\n'
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Str", ['r"abc"'])

    @testutils.only_for_versions_higher("3.6")
    def test_handling_format_strings_basic(self):
        source = '1 + f"abc{a}"\n'
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("JoinedStr", ['f"', "abc", "FormattedValue", "", '"'])
        checker.check_children("FormattedValue", ["{", "", "Name", "", "}"])

    @testutils.only_for_versions_higher("3.6")
    def test_handling_format_strings_with_implicit_join(self):
        source = '''"1" + rf'abc{a}' f"""xxx{b} """\n'''
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "JoinedStr",
            ["rf'", "abc", "FormattedValue", '\' f"""xxx', "FormattedValue", " ", '"""'],
        )
        checker.check_children("FormattedValue", ["{", "", "Name", "", "}"])

    @testutils.only_for_versions_higher("3.6")
    def test_handling_format_strings_with_format_spec(self):
        source = 'f"abc{a:01}"\n'
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("JoinedStr", ['f"', "abc", "FormattedValue", "", '"'])
        checker.check_children(
            "FormattedValue", ["{", "", "Name", "", ":", "", "01", "", "}"]
        )

    @testutils.only_for_versions_higher("3.6")
    def test_handling_format_strings_with_inner_format_spec(self):
        source = 'f"abc{a:{length}01}"\n'
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("JoinedStr", ['f"', "abc", "FormattedValue", "", '"'])
        checker.check_children(
            "FormattedValue",
            ["{", "", "Name", "", ":", "{", "Name", "}", "01", "", "}"],
        )

    @testutils.only_for_versions_higher("3.6")
    def test_handling_format_strings_with_expression(self):
        source = 'f"abc{a + b}"\n'
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("JoinedStr", ['f"', "abc", "FormattedValue", "", '"'])
        checker.check_children("FormattedValue", ["{", "", "BinOp", "", "}"])

    def test_complex_number_literals(self):
        source = "1.0e2j + a"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("BinOp", ["Num", " ", "+", " ", "Name"])
        checker.check_children("Num", ["1.0e2j"])

    def test_ass_attr_node(self):
        source = "a.b = 1\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Attribute", 0, source.index("=") - 1)
        checker.check_children("Attribute", ["Name", "", ".", "", "b"])

    def test_ass_list_node(self):
        source = "[a, b] = 1, 2\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("List", 0, source.index("]") + 1)
        checker.check_children("List", ["[", "", "Name", "", ",", " ", "Name", "", "]"])

    def test_ass_tuple(self):
        source = "a, b = range(2)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Tuple", 0, source.index("=") - 1)
        checker.check_children("Tuple", ["Name", "", ",", " ", "Name"])

    def test_ass_tuple2(self):
        source = "(a, b) = range(2)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Tuple", 0, source.index("=") - 1)
        checker.check_children(
            "Tuple", ["(", "", "Name", "", ",", " ", "Name", "", ")"]
        )

    def test_assert(self):
        source = "assert True\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Assert", 0, len(source) - 1)
        checker.check_children("Assert", ["assert", " ", NameConstant])

    def test_assert2(self):
        source = 'assert True, "error"\n'
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Assert", 0, len(source) - 1)
        checker.check_children(
            "Assert", ["assert", " ", NameConstant, "", ",", " ", "Str"]
        )

    def test_aug_assign_node(self):
        source = "a += 1\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("AugAssign", 0, len(source) - 1)
        checker.check_children("AugAssign", ["Name", " ", "+", "", "=", " ", "Num"])

    def test_bitand(self):
        source = "1 & 2\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("BinOp", 0, len(source) - 1)
        checker.check_children("BinOp", ["Num", " ", "&", " ", "Num"])

    def test_bitor(self):
        source = "1 | 2\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("BinOp", ["Num", " ", "|", " ", "Num"])

    def test_call_func(self):
        source = "f(1, 2)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Call", 0, len(source) - 1)
        checker.check_children(
            "Call", ["Name", "", "(", "", "Num", "", ",", " ", "Num", "", ")"]
        )

    def test_call_func_and_keywords(self):
        source = "f(1, p=2)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call", ["Name", "", "(", "", "Num", "", ",", " ", "keyword", "", ")"]
        )

    @testutils.only_for_versions_lower("3.5")
    def test_call_func_and_star_args(self):
        source = "f(1, *args)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call", ["Name", "", "(", "", "Num", "", ",", " ", "*", "", "Name", "", ")"]
        )

    def test_call_func_and_star_argspython35(self):
        source = "f(1, *args)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call", ["Name", "", "(", "", "Num", "", ",", " *", "Starred", "", ")"]
        )

    @testutils.only_for_versions_lower("3.5")
    def test_call_func_and_only_dstar_args(self):
        source = "f(**kwds)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Call", ["Name", "", "(", "", "**", "", "Name", "", ")"])

    def test_call_func_and_only_dstar_args_python35(self):
        source = "f(**kwds)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Call", ["Name", "", "(", "**", "keyword", "", ")"])

    @testutils.only_for_versions_lower("3.5")
    def test_call_func_and_both_varargs_and_kwargs(self):
        source = "f(*args, **kwds)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call",
            ["Name", "", "(", "", "*", "", "Name", "", ",", " ", "**", "", "Name", "", ")"],
        )

    def test_call_func_and_both_varargs_and_kwargs_python35(self):
        source = "f(*args, **kwds)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call",
            ["Name", "", "(", "*", "Starred", "", ",", " **", "keyword", "", ")"],
        )

    def test_class_node(self):
        source = dedent('''\
            class A(object):
                """class docs"""
                pass
        ''')
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Class", 0, len(source) - 1)
        checker.check_children(
            "Class",
            ["class", " ", "A", "", "(", "", "Name", "", ")", "", ":", "\n    ", "Expr", "\n    ", "Pass"],
        )

    def test_class_with_no_bases(self):
        source = dedent("""\
            class A:
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Class", 0, len(source) - 1)
        checker.check_children("Class", ["class", " ", "A", "", ":", "\n    ", "Pass"])

    def test_simple_compare(self):
        source = "1 < 2\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Compare", 0, len(source) - 1)
        checker.check_children("Compare", ["Num", " ", "<", " ", "Num"])

    def test_multiple_compare(self):
        source = "1 < 2 <= 3\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Compare", 0, len(source) - 1)
        checker.check_children(
            "Compare", ["Num", " ", "<", " ", "Num", " ", "<=", " ", "Num"]
        )

    def test_decorators_node(self):
        source = dedent("""\
            @d
            def f():
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("FunctionDef", 0, len(source) - 1)
        checker.check_children(
            "FunctionDef",
            ["@", "", "Name", "\n", "def", " ", "f", "", "(", "", "arguments", "", ")", "", ":", "\n    ", "Pass"],
        )

    def test_decorators_for_classes(self):
        source = dedent("""\
            @d
            class C(object):
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("ClassDef", 0, len(source) - 1)
        checker.check_children(
            "ClassDef",
            ["@", "", "Name", "\n", "class", " ", "C", "", "(", "", "Name", "", ")", "", ":", "\n    ", "Pass"],
        )

    def test_both_varargs_and_kwargs(self):
        source = dedent("""\
            def f(*args, **kwds):
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "arguments", ["*", "", "args", "", ",", " ", "**", "", "kwds"]
        )

    def test_function_node(self):
        source = dedent("""\
            def f():
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Function", 0, len(source) - 1)
        checker.check_children(
            "Function",
            ["def", " ", "f", "", "(", "", "arguments", "", ")", "", ":", "\n    ", "Pass"],
        )

    @testutils.only_for_versions_higher("3.5")
    def test_async_function_node(self):
        source = dedent("""\
            async def f():
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("AsyncFunction", 0, len(source) - 1)
        checker.check_children(
            "AsyncFunction",
            ["async", " ", "def", " ", "f", "", "(", "", "arguments", "", ")", "", ":", "\n    ", "Pass"],
        )

    def test_function_node2(self):
        source = dedent('''\
            def f(p1, **p2):
                """docs"""
                pass
        ''')
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Function", 0, len(source) - 1)
        checker.check_children(
            "Function",
            ["def", " ", "f", "", "(", "", "arguments", "", ")", "", ":", "\n    ", "Expr", "\n    ", "Pass"],
        )
        expected_child = ast.arg.__name__
        checker.check_children(
            "arguments", [expected_child, "", ",", " ", "**", "", "p2"]
        )

    def test_dict_node(self):
        source = "{1: 2, 3: 4}\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Dict", 0, len(source) - 1)
        checker.check_children(
            "Dict",
            ["{", "", "Num", "", ":", " ", "Num", "", ",", " ", "Num", "", ":", " ", "Num", "", "}"],
        )

    def test_dict_node_with_unpacking(self):
        source = "{**dict1, **dict2}\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Dict", 0, len(source) - 1)
        checker.check_children(
            "Dict", ["{", "", "**", "", "Name", "", ",", " ", "**", "", "Name", "", "}"]
        )

    def test_div_node(self):
        source = "1 / 2\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("BinOp", 0, len(source) - 1)
        checker.check_children("BinOp", ["Num", " ", "/", " ", "Num"])

    def test_for_node(self):
        source = dedent("""\
            for i in range(1):
                pass
            else:
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("For", 0, len(source) - 1)
        checker.check_children(
            "For",
            ["for", " ", "Name", " ", "in", " ", "Call", "", ":", "\n    ", "Pass", "\n", "else", "", ":", "\n    ", "Pass"],
        )

    @testutils.only_for_versions_higher("3.5")
    def test_async_for_node(self):
        source = dedent("""\
            async def foo():
                async for i in range(1):
                    pass
                else:
                    pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("AsyncFor", source.index("async for"), len(source) - 1)
        checker.check_children(
            "AsyncFor",
            ["async", " ", "for", " ", "Name", " ", "in", " ", "Call", "", ":", "\n        ", "Pass", "\n    ", "else", "", ":", "\n        ", "Pass"],
        )

    @testutils.only_for_versions_higher("3.8")
    def test_named_expr_node(self):
        source = dedent("""\
            if a := 10 == 10:
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        start = source.index("a")
        checker.check_region("NamedExpr", start, start + 13)
        checker.check_children("NamedExpr", ["Name", " ", ":=", " ", "Compare"])

    def test_normal_from_node(self):
        source = "from x import y\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("ImportFrom", 0, len(source) - 1)
        checker.check_children(
            "ImportFrom", ["from", " ", "x", " ", "import", " ", "alias"]
        )
        checker.check_children("alias", ["y"])

    def test_from_node(self):
        source = "from ..x import y as z\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("ImportFrom", 0, len(source) - 1)
        checker.check_children(
            "ImportFrom", ["from", " ", ".", "", ".", "", "x", " ", "import", " ", "alias"]
        )
        checker.check_children("alias", ["y", " ", "as", " ", "z"])

    def test_from_node_relative_import(self):
        source = "from . import y as z\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("ImportFrom", 0, len(source) - 1)
        checker.check_children(
            "ImportFrom", ["from", " ", ".", " ", "import", " ", "alias"]
        )
        checker.check_children("alias", ["y", " ", "as", " ", "z"])

    def test_from_node_whitespace_around_dots_1(self):
        source = "from . . . import y as z\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("ImportFrom", 0, len(source) - 1)
        checker.check_children(
            "ImportFrom", ["from", " ", ".", " ", ".", " ", ".", " ", "import", " ", "alias"]
        )
        checker.check_children("alias", ["y", " ", "as", " ", "z"])

    def test_from_node_whitespace_around_dots_2(self):
        source = "from . a . b import y as z\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("ImportFrom", 0, len(source) - 1)
        checker.check_children(
            "ImportFrom", ["from", " ", ".", " ", "a", " . ", "b", " ", "import", " ", "alias"]
        )
        checker.check_children("alias", ["y", " ", "as", " ", "z"])

    def test_simple_gen_expr_node(self):
        source = "zip(i for i in x)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("GeneratorExp", 4, len(source) - 2)
        checker.check_children("GeneratorExp", ["Name", " ", "comprehension"])
        checker.check_children(
            "comprehension", ["for", " ", "Name", " ", "in", " ", "Name"]
        )

    def test_gen_expr_node_handling_surrounding_parens(self):
        source = "(i for i in x)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("GeneratorExp", 0, len(source) - 1)
        checker.check_children(
            "GeneratorExp", ["(", "", "Name", " ", "comprehension", "", ")"]
        )

    def test_gen_expr_node2(self):
        source = "zip(i for i in range(1) if i == 1)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "comprehension",
            ["for", " ", "Name", " ", "in", " ", "Call", " ", "if", " ", "Compare"],
        )

    def test_get_attr_node(self):
        source = "a.b\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Attribute", 0, len(source) - 1)
        checker.check_children("Attribute", ["Name", "", ".", "", "b"])

    def test_global_node(self):
        source = "global a, b\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Global", 0, len(source) - 1)
        checker.check_children("Global", ["global", " ", "a", "", ",", " ", "b"])

    def test_if_node(self):
        source = "if True:\n    pass\nelse:\n    pass\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("If", 0, len(source) - 1)
        checker.check_children(
            "If",
            ["if", " ", NameConstant, "", ":", "\n    ", "Pass", "\n", "else", "", ":", "\n    ", "Pass"],
        )

    def test_if_node2(self):
        source = dedent("""\
            if True:
                pass
            elif False:
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("If", 0, len(source) - 1)
        checker.check_children(
            "If", ["if", " ", NameConstant, "", ":", "\n    ", "Pass", "\n", "If"]
        )

    def test_if_node3(self):
        source = dedent("""\
            if True:
                pass
            else:
                if True:
                    pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("If", 0, len(source) - 1)
        checker.check_children(
            "If",
            ["if", " ", NameConstant, "", ":", "\n    ", "Pass", "\n", "else", "", ":", "\n    ", "If"],
        )

    def test_import_node(self):
        source = "import a, b as c\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Import", 0, len(source) - 1)
        checker.check_children(
            "Import", ["import", " ", "alias", "", ",", " ", "alias"]
        )

    def test_import_node_whitespace_around_dots(self):
        source = "import a . b, b as c\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Import", 0, len(source) - 1)
        checker.check_children(
            "Import", ["import", " ", "alias", "", ",", " ", "alias"]
        )

    def test_lambda_node(self):
        source = "lambda a, b=1, *z: None\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Lambda", 0, len(source) - 1)
        checker.check_children(
            "Lambda", ["lambda", " ", "arguments", "", ":", " ", NameConstant]
        )
        expected_child = ast.arg.__name__
        checker.check_children(
            "arguments",
            [expected_child, "", ",", " ", expected_child, "", "=", "", "Num", "", ",", " ", "*", "", "z"],
        )

    def test_list_node(self):
        source = "[1, 2]\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("List", 0, len(source) - 1)
        checker.check_children("List", ["[", "", "Num", "", ",", " ", "Num", "", "]"])

    def test_list_comp_node(self):
        source = "[i for i in range(1) if True]\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("ListComp", 0, len(source) - 1)
        checker.check_children(
            "ListComp", ["[", "", "Name", " ", "comprehension", "", "]"]
        )
        checker.check_children(
            "comprehension",
            ["for", " ", "Name", " ", "in", " ", "Call", " ", "if", " ", NameConstant],
        )

    def test_list_comp_node_with_multiple_comprehensions(self):
        source = "[i for i in range(1) for j in range(1) if True]\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("ListComp", 0, len(source) - 1)
        checker.check_children(
            "ListComp",
            ["[", "", "Name", " ", "comprehension", " ", "comprehension", "", "]"],
        )
        checker.check_children(
            "comprehension",
            ["for", " ", "Name", " ", "in", " ", "Call", " ", "if", " ", NameConstant],
        )

    def test_set_node(self):
        source = "{1, 2}\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Set", 0, len(source) - 1)
        checker.check_children("Set", ["{", "", "Num", "", ",", " ", "Num", "", "}"])

    def test_set_comp_node(self):
        source = "{i for i in range(1) if True}\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("SetComp", 0, len(source) - 1)
        checker.check_children(
            "SetComp", ["{", "", "Name", " ", "comprehension", "", "}"]
        )
        checker.check_children(
            "comprehension",
            ["for", " ", "Name", " ", "in", " ", "Call", " ", "if", " ", NameConstant],
        )

    def test_dict_comp_node(self):
        source = "{i:i for i in range(1) if True}\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("DictComp", 0, len(source) - 1)
        checker.check_children(
            "DictComp",
            ["{", "", "Name", "", ":", "", "Name", " ", "comprehension", "", "}"],
        )
        checker.check_children(
            "comprehension",
            ["for", " ", "Name", " ", "in", " ", "Call", " ", "if", " ", NameConstant],
        )

    def test_ext_slice_node(self):
        source = "x = xs[0,:]\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        if sys.version_info >= (3, 9):
            checker.check_region("Tuple", 7, len(source) - 2)
            checker.check_children("Tuple", ["Num", "", ",", "", "Slice"])
        else:
            checker.check_region("ExtSlice", 7, len(source) - 2)
            checker.check_children("ExtSlice", ["Index", "", ",", "", "Slice"])

    def test_simple_module_node(self):
        source = "pass\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Module", 0, len(source))
        checker.check_children("Module", ["", "Pass", "\n"])

    def test_module_node(self):
        source = dedent('''\
            """docs"""
            pass
        ''')
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Module", 0, len(source))
        checker.check_children("Module", ["", "Expr", "\n", "Pass", "\n"])
        checker.check_children("Str", ['"""docs"""'])

    def test_not_and_or_nodes(self):
        source = "not True or False\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Expr", ["BoolOp"])
        checker.check_children("BoolOp", ["UnaryOp", " ", "or", " ", NameConstant])

    def test_raise_node_bare(self):
        source = "raise\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Raise", 0, len(source) - 1)
        checker.check_children("Raise", ["raise"])

    def test_raise_node_for_python3(self):
        source = "raise x(y)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Raise", 0, len(source) - 1)
        checker.check_children("Raise", ["raise", " ", "Call"])

    def test_raise_node_for_python3_with_cause(self):
        source = "raise x(y) from e\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_region("Raise", 0, len(source) - 1)
        checker.check_children("Raise", ["raise", " ", "Call", " ", "from", " ", "Name"])

    def test_return_node(self):
        source = dedent("""\
            def f():
                return None
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Return", ["return", " ", NameConstant])

    def test_empty_return_node(self):
        source = dedent("""\
            def f():
                return
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Return", ["return"])

    def test_simple_slice_node(self):
        source = "a[1:2]\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Subscript", ["Name", "", "[", "", "Slice", "", "]"])
        checker.check_children("Slice", ["Num", "", ":", "", "Num"])

    def test_slice_node2(self):
        source = "a[:]\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Subscript", ["Name", "", "[", "", "Slice", "", "]"])
        checker.check_children("Slice", [":"])

    def test_simple_subscript(self):
        source = "a[1]\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        if sys.version_info >= (3, 9):
            checker.check_children("Subscript", ["Name", "", "[", "", "Num", "", "]"])
        else:
            checker.check_children("Subscript", ["Name", "", "[", "", "Index", "", "]"])
            checker.check_children("Index", ["Num"])

    def test_tuple_node(self):
        source = "(1, 2)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Tuple", ["(", "", "Num", "", ",", " ", "Num", "", ")"])

    def test_tuple_node2(self):
        source = "#(\n1, 2\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Tuple", ["Num", "", ",", " ", "Num"])

    def test_tuple_with_complex_parentheses1(self):
        source = "a = ( # (he\n ((((), None))))\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Tuple", ["(", "", "Tuple", "", ",", " ", NameConstant, "", ")"]
        )

    def test_tuple_with_complex_parentheses2(self):
        source = "a = ( # (he\n ((((('a')), ('b')))))\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Tuple", ["(", "", "((", "Str", "))", ",", " (", "Str", ")", "", ")"]
        )

    def test_tuple_with_complex_parentheses3(self):
        source = "a = ((), (([],), []),)"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Tuple", ["(", "", "Tuple", "", ",", " ", "Tuple", ",", ")"]
        )

    def test_one_item_tuple_node(self):
        source = "(1,)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Tuple", ["(", "", "Num", ",", ")"])

    def test_empty_tuple_node(self):
        source = "()\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Tuple", ["()"])

    def test_empty_tuple_node2(self):
        source = "a = ((), None)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Tuple", ["(", "", "Tuple", "", ",", " ", NameConstant, "", ")"]
        )

    def test_empty_tuple_node3(self):
        source = "a = (), None\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Tuple", ["Tuple", "", ",", " ", NameConstant]
        )

    def test_yield_node(self):
        source = dedent("""\
            def f():
                yield None
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Yield", ["yield", " ", NameConstant])

    @testutils.only_for_versions_higher("3.3")
    def test_yield_from_node(self):
        source = dedent("""\
            def f(lst):
                yield from lst
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("YieldFrom", ["yield", " ", "from", " ", "Name"])

    def test_while_node(self):
        source = dedent("""\
            while True:
                pass
            else:
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "While",
            ["while", " ", NameConstant, "", ":", "\n    ", "Pass", "\n", "else", "", ":", "\n    ", "Pass"],
        )

    def test_with_node(self):
        source = dedent("""\
            from __future__ import with_statement
            with a as b:
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "With",
            ["with", " ", "Name", " ", "as", " ", "Name", "", ":", "\n    ", "Pass"],
        )

    def test_async_with_node(self):
        source = dedent("""\
            async def afunc():
                async with a as b:
                    pass\n
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "AsyncWith",
            ["async", " ", "with", " ", "Name", " ", "as", " ", "Name", "", ":", "\n        ", "Pass"],
        )

    def test_try_finally_node(self):
        source = dedent("""\
            try:
                pass
            finally:
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        node_to_test = "Try"
        expected_children = ["try", "", ":", "\n    ", "Pass", "\n", "finally", "", ":", "\n    ", "Pass"]
        checker.check_children(node_to_test, expected_children)

    def test_try_except_node(self):
        source = dedent("""\
            try:
                pass
            except Exception as e:
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        node_to_test = "Try"
        checker.check_children(
            node_to_test,
            ["try", "", ":", "\n    ", "Pass", "\n", ("excepthandler", "ExceptHandler")],
        )
        expected_child = "e"
        checker.check_children(
            ("excepthandler", "ExceptHandler"),
            ["except", " ", "Name", " ", "as", " ", expected_child, "", ":", "\n    ", "Pass"],
        )

    def test_try_except_and_finally_node(self):
        source = dedent("""\
            try:
                pass
            except:
                pass
            finally:
                pass
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        node_to_test = "Try"
        expected_children = ["try", "", ":", "\n    ", "Pass", "\n", "ExceptHandler", "\n", "finally", "", ":", "\n    ", "Pass"]
        checker.check_children(node_to_test, expected_children)

    def test_ignoring_comments(self):
        source = "#1\n1\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        start = source.rindex("1")
        checker.check_region("Num", start, start + 1)

    def test_simple_sliceobj(self):
        source = "a[1::3]\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Slice", ["Num", "", ":", "", ":", "", "Num"])

    def test_ignoring_strings_that_start_with_a_char(self):
        source = 'r"""("""\n1\n'
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Module", ["", "Expr", "\n", "Expr", "\n"])

    def test_semicolon(self):
        source = "1;\n"
        patchedast.get_patched_ast(source, True)

    def test_if_exp_node(self):
        source = "1 if True else 2\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "IfExp", ["Num", " ", "if", " ", NameConstant, " ", "else", " ", "Num"]
        )

    def test_delete_node(self):
        source = "del a, b\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Delete", ["del", " ", "Name", "", ",", " ", "Name"])

    @testutils.only_for_versions_lower("3.5")
    def test_starargs_before_keywords_legacy(self):
        source = "foo(*args, a=1)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call",
            ["Name", "", "(", "", "*", "", "Name", "", ",", " ", "keyword", "", ")"],
        )

    @testutils.only_for_versions_lower("3.5")
    def test_starargs_in_keywords_legacy(self):
        source = "foo(a=1, *args, b=2)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call",
            ["Name", "", "(", "", "keyword", "", ",", " ", "*", "", "Name", "", ",", " ", "keyword", "", ")"],
        )

    @testutils.only_for_versions_lower("3.5")
    def test_starargs_after_keywords_legacy(self):
        source = "foo(a=1, *args)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call",
            ["Name", "", "(", "", "keyword", "", ",", " ", "*", "", "Name", "", ")"],
        )

    def test_starargs_before_keywords(self):
        source = "foo(*args, a=1)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call", ["Name", "", "(", "*", "Starred", "", ",", " ", "keyword", "", ")"]
        )

    def test_starargs_in_keywords(self):
        source = "foo(a=1, *args, b=2)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call",
            ["Name", "", "(", "", "keyword", "", ",", " *", "Starred", "", ",", " ", "keyword", "", ")"],
        )

    def test_starargs_in_positional(self):
        source = "foo(a, *b, c)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call",
            ["Name", "", "(", "", "Name", "", ",", " *", "Starred", "", ",", " ", "Name", "", ")"],
        )

    def test_starargs_after_keywords(self):
        source = "foo(a=1, *args)\n"
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children(
            "Call", ["Name", "", "(", "", "keyword", "", ",", " *", "Starred", "", ")"]
        )

    @testutils.only_for_versions_higher("3.5")
    def test_await_node(self):
        source = dedent("""\
            async def f():
                await sleep()
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Await", ["await", " ", "Call"])

    @testutils.only_for_versions_higher("3.10")
    def test_match_node_with_constant_match_value(self):
        source = dedent("""\
            match x:
                case 1:
                    print(x)
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        self.assert_single_case_match_block(checker, "MatchValue")
        checker.check_children("MatchValue", [
            "Num"
        ])

    @testutils.only_for_versions_higher("3.10")
    def test_match_node_match_case_with_guard(self):
        source = dedent("""\
            match x:
                case int(n) if x < 10:
                    print(n)
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        checker.check_children("Match", [
            "match",
            " ",
            "Name",
            "",
            ":",
            "\n    ",
            "match_case",
        ])
        checker.check_children("match_case", [
            "case",
            " ",
            "MatchClass",
            " ",
            "if",
            " ",
            "Compare",
            "",
            ":",
            "\n        ",
            "Expr",
        ])

    @testutils.only_for_versions_higher("3.10")
    def test_match_node_with_match_class(self):
        source = dedent("""\
            match x:
                case Foo(1):
                    print(x)
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        self.assert_single_case_match_block(checker, "MatchClass")
        checker.check_children("MatchClass", [
            "Name",
            "",
            "(",
            "",
            "MatchValue",
            "",
            ")",
        ])
        checker.check_children("MatchValue", [
            "Num"
        ])

    @testutils.only_for_versions_higher("3.10")
    def test_match_node_with_wildcard(self):
        source = dedent("""\
            match x:
                case _:
                    print(x)
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        self.assert_single_case_match_block(checker, "MatchAs")
        checker.check_children("MatchAs", [
            "_"
        ])

    @testutils.only_for_versions_higher("3.10")
    def test_match_node_with_match_as_capture_pattern(self):
        source = dedent("""\
            match x:
                case myval:
                    print(myval)
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        self.assert_single_case_match_block(checker, "MatchAs")
        checker.check_children("MatchAs", [
            "myval"
        ])

    @testutils.only_for_versions_higher("3.10")
    def test_match_node_with_match_as_capture_pattern_with_explicit_name(self):
        source = dedent("""\
            match x:
                case "foo" as myval:
                    print(myval)
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        self.assert_single_case_match_block(checker, "MatchAs")
        checker.check_children("MatchAs", [
            "MatchValue",
            " ",
            "as",
            " ",
            "myval",
        ])

    @testutils.only_for_versions_higher("3.10")
    def test_match_node_with_match_class_simple_match_as_capture_pattern(self):
        source = dedent("""\
            match x:
                case Foo(x):
                    print(x)
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        self.assert_single_case_match_block(checker, "MatchClass")
        checker.check_children("MatchClass", [
            "Name",
            "",
            "(",
            "",
            "MatchAs",
            "",
            ")",
        ])

    @testutils.only_for_versions_higher("3.10")
    def test_match_node_with_match_class_named_argument(self):
        source = dedent("""\
            match x:
                case Foo(x=10, y="20"):
                    print(x)
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        self.assert_single_case_match_block(checker, "MatchClass")
        checker.check_children("MatchClass", [
            "Name",
            "",
            "(",
            "",
            "x",
            "",
            "=",
            "",
            "MatchValue",
            "",
            ",",
            " ",
            "y",
            "",
            "=",
            "",
            "MatchValue",
            "",
            ")",
        ])

    @testutils.only_for_versions_higher("3.10")
    def test_match_node_with_match_class_match_as_capture_pattern_with_explicit_name(self):
        source = dedent("""\
            match x:
                case Foo(x) as b:
                    print(x)
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        self.assert_single_case_match_block(checker, "MatchAs")
        checker.check_children("MatchAs", [
            "MatchClass",
            " ",
            "as",
            " ",
            "b",
        ])
        checker.check_children("MatchClass", [
            "Name",
            "",
            "(",
            "",
            "MatchAs",
            "",
            ")",
        ])

    @testutils.only_for_versions_higher("3.10")
    def test_match_node_with_match_mapping_match_as(self):
        source = dedent("""\
            match x:
                case {"a": b} as c:
                    print(x)
        """)
        ast_frag = patchedast.get_patched_ast(source, True)
        checker = _ResultChecker(self, ast_frag)
        self.assert_single_case_match_block(checker, "MatchAs")
        checker.check_children("MatchAs", [
            "MatchMapping",
            " ",
            "as",
            " ",
            "c",
        ])
        checker.check_children("MatchMapping", [
            "{",
            "",
            "Str",
            "",
            ":",
            " ",
            "MatchAs",
            "",
            "}",
        ])


class _ResultChecker:
    def __init__(self, test_case, ast):
        self.test_case = test_case
        self.ast = ast

    def check_region(self, text, start, end):
        node = self._find_node(text)
        if node is None:
            self.test_case.fail("Node <%s> cannot be found" % text)
        self.test_case.assertEqual((start, end), node.region)

    def _find_node(self, text):
        """
        Find the node in `self.ast` whose type is named in `text`.

        :param text: ast node name

        Generally, the test should only have a single matching node, as it make
        the test much harder to understand when there may be multiple matches.

        If `self.ast` contains more than one nodes that matches `text`, then
        the **outer-most last match** takes precedence.

        For example, given that we are looking for `ast.Call` node:

            checker._find_node("Call")

        and given that `self.ast` is the AST for this code:

            func_a(1, func_b(2, 3)) + func_c(4, func_d(5, 6))

        the outer-most last match would be the ast node representing this bit:

            func_c(4, func_d(5, 6))

        Note that the order of traversal is based on the order of ast nodes,
        which usually, but not always, match textual order.
        """
        goal = text
        if not isinstance(text, (tuple, list)):
            goal = [text]

        class Search:
            result = None

            def __call__(self, node):
                for text in goal:
                    if str(node).startswith(text):
                        self.result = node
                        break
                    if ast.get_node_type_name(node).startswith(text):
                        self.result = node
                        break
                return self.result is not None

        search = Search()
        ast.call_for_nodes(self.ast, search)
        return search.result

    def check_children(self, text, children):
        node = self._find_node(text)
        if node is None:
            self.test_case.fail("Node <%s> cannot be found" % text)
        result = list(node.sorted_children)
        self.test_case.assertEqual(len(children), len(result))
        for expected, child in zip(children, result):
            goals = expected
            if not isinstance(expected, (tuple, list)):
                goals = [expected]
            for goal in goals:
                if goal == "" or isinstance(child, (str, bytes)):
                    self.test_case.assertEqual(goal, child)
                    break
            else:
                self.test_case.assertNotEqual("", text, "probably ignoring some node")
                self.test_case.assertTrue(
                    ast.get_node_type_name(child).startswith(expected),
                    msg="Expected <%s> but was <%s>"
                    % (expected, ast.get_node_type_name(child)),
                )



from rope.base import utils
from rope.base.oi import objectdb
from rope.base.serializer import json_to_python, python_to_json


class MemoryDB(objectdb.FileDict):
    def __init__(self, project, persist=None):
        self.project = project
        self._persist = persist
        self.files = self
        self._load_files()
        self.project.data_files.add_write_hook(self.write)

    def _load_files(self):
        self._files = {}
        if self.persist:
            result = self.project.data_files.read_data("objectdb")
            if result is not None:
                self._files = result

    def keys(self):
        return self._files.keys()

    def __iter__(self):
        yield from self._files

    def __len__(self):
        return len(self._files)

    def __setitem__(self):
        raise NotImplementedError()

    def __contains__(self, key):
        return key in self._files

    def __getitem__(self, key):
        return FileInfo(self._files[key])

    def create(self, path):
        self._files[path] = {}

    def rename(self, file, newfile):
        if file not in self._files:
            return
        self._files[newfile] = self._files[file]
        del self[file]

    def __delitem__(self, file):
        del self._files[file]

    def write(self):
        if self.persist:
            self.project.data_files.write_data("objectdb", self._files)

    @property
    @utils.deprecated("compress_objectdb is no longer supported")
    def compress(self):
        return False

    @property
    def persist(self):
        if self._persist is not None:
            return self._persist
        else:
            return self.project.prefs.get("save_objectdb", False)


class FileInfo(objectdb.FileInfo):
    def __init__(self, scopes):
        self.scopes = scopes

    def create_scope(self, key):
        self.scopes[key] = ScopeInfo()

    def keys(self):
        return self.scopes.keys()

    def __contains__(self, key):
        return key in self.scopes

    def __getitem__(self, key):
        return self.scopes[key]

    def __delitem__(self, key):
        del self.scopes[key]

    def __iter__(self):
        yield from self.scopes

    def __len__(self):
        return len(self.scopes)

    def __setitem__(self):
        raise NotImplementedError()


class ScopeInfo(objectdb.ScopeInfo):
    def __init__(self):
        self.call_info = {}
        self.per_name = {}

    def get_per_name(self, name):
        return self.per_name.get(name, None)

    def save_per_name(self, name, value):
        self.per_name[name] = value

    def get_returned(self, parameters):
        return self.call_info.get(parameters, None)

    def get_call_infos(self):
        for args, returned in self.call_info.items():
            yield objectdb.CallInfo(args, returned)

    def add_call(self, parameters, returned):
        self.call_info[parameters] = returned

    def __getstate__(self):
        original_data = (self.call_info, self.per_name)
        encoded = python_to_json(original_data, version=2)
        encoded["$"] = "ScopeInfo"
        return encoded

    def __setstate__(self, data):
        if isinstance(data, tuple) and len(data) == 2:
            # legacy pickle-based serialization
            self.call_info, self.per_name = data
        else:
            # new serialization
            assert data["$"] == "ScopeInfo"
            self.call_info, self.per_name = json_to_python(data)



from rope.base.oi.type_hinting import evaluate
from rope.base.oi.type_hinting.resolvers import interfaces


class Resolver(interfaces.IResolver):
    def __call__(self, hint, pyobject):
        """
        :param hint: For example "List[int]" or "(Foo, Bar) -> Baz" or simple "Foo"
        :type hint: str
        :type pyobject: rope.base.pyobjects.PyDefinedObject | rope.base.pyobjects.PyObject
        :rtype: rope.base.pyobjects.PyDefinedObject | rope.base.pyobjects.PyObject or None
        """
        try:
            return evaluate.evaluate(hint, pyobject)
        except (Exception):
            pass



from rope.base import codeanalyze


def get_indents(lines, lineno):
    return codeanalyze.count_line_indents(lines.get_line(lineno))


def find_minimum_indents(source_code):
    result = 80
    lines = source_code.split("\n")
    for line in lines:
        if line.strip() == "":
            continue
        result = min(result, codeanalyze.count_line_indents(line))
    return result


def indent_lines(source_code, amount):
    if amount == 0:
        return source_code
    lines = source_code.splitlines(True)
    result = []
    for line in lines:
        if line.strip() == "":
            result.append("\n")
            continue
        if amount < 0:
            indents = codeanalyze.count_line_indents(line)
            result.append(max(0, indents + amount) * " " + line.lstrip())
        else:
            result.append(" " * amount + line)
    return "".join(result)


def fix_indentation(code: str, new_indents: int) -> str:
    """Change the indentation of `code` to `new_indents`"""
    min_indents = find_minimum_indents(code)
    return indent_lines(code, new_indents - min_indents)


def add_methods(pymodule, class_scope, methods_sources):
    source_code = pymodule.source_code
    lines = pymodule.lines
    insertion_line = class_scope.get_end()
    if class_scope.get_scopes():
        insertion_line = class_scope.get_scopes()[-1].get_end()
    insertion_offset = lines.get_line_end(insertion_line)
    methods = "\n\n" + "\n\n".join(methods_sources)
    indented_methods = fix_indentation(
        methods,
        get_indents(lines, class_scope.get_start())
        + get_indent(pymodule.pycore.project),
    )
    result = []
    result.append(source_code[:insertion_offset])
    result.append(indented_methods)
    result.append(source_code[insertion_offset:])
    return "".join(result)


def get_body(pyfunction):
    """Return unindented function body"""
    # FIXME scope = pyfunction.get_scope()
    pymodule = pyfunction.get_module()
    start, end = get_body_region(pyfunction)
    return fix_indentation(pymodule.source_code[start:end], 0)


def get_body_region(defined):
    """Return the start and end offsets of function body"""
    scope = defined.get_scope()
    pymodule = defined.get_module()
    lines = pymodule.lines
    node = defined.get_ast()
    start_line = node.lineno
    if defined.get_doc() is None:
        start_line = node.body[0].lineno
    elif len(node.body) > 1:
        start_line = node.body[1].lineno
    start = lines.get_line_start(start_line)
    scope_start = pymodule.logical_lines.logical_line_in(scope.start)
    if scope_start[1] >= start_line:
        # a one-liner!
        # XXX: what if colon appears in a string
        start = pymodule.source_code.index(":", start) + 1
        while pymodule.source_code[start].isspace():
            start += 1
    end = min(lines.get_line_end(scope.end) + 1, len(pymodule.source_code))
    return start, end


def get_indent(project):
    return project.prefs.get("indent_size", 4)



def __rope_start_everything():
    import os
    import socket
    import sys

    try:
        import cPickle as pickle
    except ImportError:
        import pickle
    import base64
    import hashlib
    import hmac
    import inspect
    import marshal
    import threading
    import types

    class _MessageSender:
        def send_data(self, data):
            pass

    class _SocketSender(_MessageSender):
        def __init__(self, port, key):
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(("127.0.0.1", port))
            self.my_file = s.makefile("wb")
            self.key = base64.b64decode(key)

        def send_data(self, data):
            if not self.my_file.closed:
                pickled_data = base64.b64encode(
                    pickle.dumps(data, pickle.HIGHEST_PROTOCOL)
                )
                dgst = hmac.new(self.key, pickled_data, hashlib.sha256).digest()
                self.my_file.write(base64.b64encode(dgst) + b":" + pickled_data + b"\n")

        def close(self):
            self.my_file.close()

    class _FileSender(_MessageSender):
        def __init__(self, file_name):
            self.my_file = open(file_name, "wb")

        def send_data(self, data):
            if not self.my_file.closed:
                marshal.dump(data, self.my_file)

        def close(self):
            self.my_file.close()

    def _cached(func):
        cache = {}

        def newfunc(self, arg):
            if arg in cache:
                return cache[arg]
            result = func(self, arg)
            cache[arg] = result
            return result

        return newfunc

    class _FunctionCallDataSender:
        def __init__(self, send_info, project_root):
            self.project_root = project_root
            if send_info[0].isdigit():
                port, key = send_info.split(":", 1)
                self.sender = _SocketSender(int(port), key)
            else:
                self.sender = _FileSender(send_info)

            def global_trace(frame, event, arg):
                # HACK: Ignoring out->in calls
                # This might lose some information
                if self._is_an_interesting_call(frame):
                    return self.on_function_call

            sys.settrace(global_trace)
            threading.settrace(global_trace)

        def on_function_call(self, frame, event, arg):
            if event != "return":
                return
            args = []
            returned = ("unknown",)
            code = frame.f_code
            for argname in code.co_varnames[: code.co_argcount]:
                try:
                    argvalue = self._object_to_persisted_form(frame.f_locals[argname])
                    args.append(argvalue)
                except (TypeError, AttributeError):
                    args.append(("unknown",))
            try:
                returned = self._object_to_persisted_form(arg)
            except (TypeError, AttributeError):
                pass
            try:
                data = (
                    self._object_to_persisted_form(frame.f_code),
                    tuple(args),
                    returned,
                )
                self.sender.send_data(data)
            except (TypeError):
                pass
            return self.on_function_call

        def _is_an_interesting_call(self, frame):
            # if frame.f_code.co_name in ['?', '<module>']:
            #    return False
            # return not frame.f_back or
            #    not self._is_code_inside_project(frame.f_back.f_code)
            if not self._is_code_inside_project(frame.f_code) and (
                not frame.f_back
                or not self._is_code_inside_project(frame.f_back.f_code)
            ):
                return False
            return True

        def _is_code_inside_project(self, code):
            source = self._path(code.co_filename)
            return (
                source is not None
                and os.path.exists(source)
                and _realpath(source).startswith(self.project_root)
            )

        @_cached
        def _get_persisted_code(self, object_):
            source = self._path(object_.co_filename)
            if not os.path.exists(source):
                raise TypeError("no source")
            return ("defined", _realpath(source), str(object_.co_firstlineno))

        @_cached
        def _get_persisted_class(self, object_):
            try:
                return (
                    "defined",
                    _realpath(inspect.getsourcefile(object_)),
                    object_.__name__,
                )
            except (TypeError, AttributeError):
                return ("unknown",)

        def _get_persisted_builtin(self, object_):
            if isinstance(object_, str):
                return ("builtin", "str")
            if isinstance(object_, list):
                holding = None
                if len(object_) > 0:
                    holding = object_[0]
                return ("builtin", "list", self._object_to_persisted_form(holding))
            if isinstance(object_, dict):
                keys = None
                values = None
                if len(object_) > 0:
                    # @todo - fix it properly, why is __locals__ being
                    # duplicated ?
                    keys = [key for key in object_.keys() if key != "__locals__"][0]
                    values = object_[keys]
                return (
                    "builtin",
                    "dict",
                    self._object_to_persisted_form(keys),
                    self._object_to_persisted_form(values),
                )
            if isinstance(object_, tuple):
                objects = []
                if len(object_) < 3:
                    for holding in object_:
                        objects.append(self._object_to_persisted_form(holding))
                else:
                    objects.append(self._object_to_persisted_form(object_[0]))
                return tuple(["builtin", "tuple"] + objects)
            if isinstance(object_, set):
                holding = None
                if len(object_) > 0:
                    for o in object_:
                        holding = o
                        break
                return ("builtin", "set", self._object_to_persisted_form(holding))
            return ("unknown",)

        def _object_to_persisted_form(self, object_):
            if object_ is None:
                return ("none",)
            if isinstance(object_, types.CodeType):
                return self._get_persisted_code(object_)
            if isinstance(object_, types.FunctionType):
                return self._get_persisted_code(object_.__code__)
            if isinstance(object_, types.MethodType):
                return self._get_persisted_code(object_.__func__.__code__)
            if isinstance(object_, types.ModuleType):
                return self._get_persisted_module(object_)
            if isinstance(object_, (str, list, dict, tuple, set)):
                return self._get_persisted_builtin(object_)
            if isinstance(object_, type):
                return self._get_persisted_class(object_)
            return ("instance", self._get_persisted_class(type(object_)))

        @_cached
        def _get_persisted_module(self, object_):
            path = self._path(object_.__file__)
            if path and os.path.exists(path):
                return ("defined", _realpath(path))
            return ("unknown",)

        def _path(self, path):
            if path.endswith(".pyc"):
                path = path[:-1]
            if path.endswith(".py"):
                return path

        def close(self):
            self.sender.close()
            sys.settrace(None)

    def _realpath(path):
        return os.path.realpath(os.path.abspath(os.path.expanduser(path)))

    send_info = sys.argv[1]
    project_root = sys.argv[2]
    file_to_run = sys.argv[3]
    run_globals = globals()
    run_globals.update(
        {"__name__": "__main__", "__builtins__": __builtins__, "__file__": file_to_run}
    )

    if send_info != "-":
        data_sender = _FunctionCallDataSender(send_info, project_root)
    del sys.argv[1:4]
    with open(file_to_run) as f:
        code = compile(f.read(), file_to_run, "exec")
        exec(code, run_globals)
    if send_info != "-":
        data_sender.close()


if __name__ == "__main__":
    __rope_start_everything()



import unittest
from textwrap import dedent

import rope.base.taskhandle
import rope.refactor.introduce_parameter
import ropetest.refactor.extracttest
import ropetest.refactor.importutilstest
import ropetest.refactor.inlinetest
import ropetest.refactor.movetest
import ropetest.refactor.multiprojecttest
import ropetest.refactor.patchedasttest
import ropetest.refactor.renametest
import ropetest.refactor.restructuretest
import ropetest.refactor.suitestest
import ropetest.refactor.usefunctiontest
from rope.base.exceptions import InterruptedTaskError, RefactoringError
from rope.refactor.encapsulate_field import EncapsulateField
from rope.refactor.introduce_factory import IntroduceFactory
from rope.refactor.localtofield import LocalToField
from rope.refactor.method_object import MethodObject
from ropetest import testutils
from ropetest.refactor import change_signature_test, similarfindertest


class MethodObjectTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, "mod")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def test_empty_method(self):
        code = dedent("""\
            def func():
                pass
        """)
        self.mod.write(code)
        replacer = MethodObject(self.project, self.mod, code.index("func"))
        expected = dedent("""\
            class _New(object):

                def __call__(self):
                    pass
        """)
        self.assertEqual(
            expected,
            replacer.get_new_class("_New"),
        )

    def test_trivial_return(self):
        code = dedent("""\
            def func():
                return 1
        """)
        self.mod.write(code)
        replacer = MethodObject(self.project, self.mod, code.index("func"))
        expected = dedent("""\
            class _New(object):

                def __call__(self):
                    return 1
        """)
        self.assertEqual(
            expected,
            replacer.get_new_class("_New"),
        )

    def test_multi_line_header(self):
        code = dedent("""\
            def func(
                ):
                return 1
        """)
        self.mod.write(code)
        replacer = MethodObject(self.project, self.mod, code.index("func"))
        expected = dedent("""\
            class _New(object):

                def __call__(self):
                    return 1
        """)
        self.assertEqual(
            expected,
            replacer.get_new_class("_New"),
        )

    def test_a_single_parameter(self):
        code = dedent("""\
            def func(param):
                return 1
        """)
        self.mod.write(code)
        replacer = MethodObject(self.project, self.mod, code.index("func"))
        expected = dedent("""\
            class _New(object):

                def __init__(self, param):
                    self.param = param

                def __call__(self):
                    return 1
        """)
        self.assertEqual(
            expected,
            replacer.get_new_class("_New"),
        )

    def test_self_parameter(self):
        code = dedent("""\
            def func(self):
                return 1
        """)
        self.mod.write(code)
        replacer = MethodObject(self.project, self.mod, code.index("func"))
        expected = dedent("""\
            class _New(object):

                def __init__(self, host):
                    self.self = host

                def __call__(self):
                    return 1
        """)
        self.assertEqual(
            expected,
            replacer.get_new_class("_New"),
        )

    def test_simple_using_passed_parameters(self):
        code = dedent("""\
            def func(param):
                return param
        """)
        self.mod.write(code)
        replacer = MethodObject(self.project, self.mod, code.index("func"))
        expected = dedent("""\
            class _New(object):

                def __init__(self, param):
                    self.param = param

                def __call__(self):
                    return self.param
        """)
        self.assertEqual(
            expected,
            replacer.get_new_class("_New"),
        )

    def test_self_keywords_and_args_parameters(self):
        code = dedent("""\
            def func(arg, *args, **kwds):
                result = arg + args[0] + kwds[arg]
                return result
        """)
        self.mod.write(code)
        replacer = MethodObject(self.project, self.mod, code.index("func"))
        expected = dedent("""\
            class _New(object):

                def __init__(self, arg, args, kwds):
                    self.arg = arg
                    self.args = args
                    self.kwds = kwds

                def __call__(self):
                    result = self.arg + self.args[0] + self.kwds[self.arg]
                    return result
        """)
        self.assertEqual(expected, replacer.get_new_class("_New"))

    def test_performing_on_not_a_function(self):
        code = dedent("""\
            my_var = 10
        """)
        self.mod.write(code)
        with self.assertRaises(RefactoringError):
            MethodObject(self.project, self.mod, code.index("my_var"))

    def test_changing_the_module(self):
        code = dedent("""\
            def func():
                return 1
        """)
        self.mod.write(code)
        replacer = MethodObject(self.project, self.mod, code.index("func"))
        self.project.do(replacer.get_changes("_New"))
        expected = dedent("""\
            def func():
                return _New()()


            class _New(object):

                def __call__(self):
                    return 1
        """)
        self.assertEqual(expected, self.mod.read())

    def test_changing_the_module_and_class_methods(self):
        code = dedent("""\
            class C(object):

                def a_func(self):
                    return 1

                def another_func(self):
                    pass
        """)
        self.mod.write(code)
        replacer = MethodObject(self.project, self.mod, code.index("func"))
        self.project.do(replacer.get_changes("_New"))
        expected = dedent("""\
            class C(object):

                def a_func(self):
                    return _New(self)()

                def another_func(self):
                    pass


            class _New(object):

                def __init__(self, host):
                    self.self = host

                def __call__(self):
                    return 1
        """)
        self.assertEqual(expected, self.mod.read())


class IntroduceFactoryTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _introduce_factory(self, resource, offset, *args, **kwds):
        factory_introducer = IntroduceFactory(self.project, resource, offset)
        changes = factory_introducer.get_changes(*args, **kwds)
        self.project.do(changes)

    def test_adding_the_method(self):
        code = dedent("""\
            class AClass(object):
                an_attr = 10
        """)
        mod = testutils.create_module(self.project, "mod")
        mod.write(code)
        expected = dedent("""\
            class AClass(object):
                an_attr = 10

                @staticmethod
                def create(*args, **kwds):
                    return AClass(*args, **kwds)
        """)
        self._introduce_factory(mod, mod.read().index("AClass") + 1, "create")
        self.assertEqual(expected, mod.read())

    def test_changing_occurrences_in_the_main_module(self):
        code = dedent("""\
            class AClass(object):
                an_attr = 10
            a_var = AClass()""")
        mod = testutils.create_module(self.project, "mod")
        mod.write(code)
        expected = dedent("""\
            class AClass(object):
                an_attr = 10

                @staticmethod
                def create(*args, **kwds):
                    return AClass(*args, **kwds)
            a_var = AClass.create()""")
        self._introduce_factory(mod, mod.read().index("AClass") + 1, "create")
        self.assertEqual(expected, mod.read())

    def test_changing_occurrences_with_arguments(self):
        code = dedent("""\
            class AClass(object):
                def __init__(self, arg):
                    pass
            a_var = AClass(10)
        """)
        mod = testutils.create_module(self.project, "mod")
        mod.write(code)
        expected = dedent("""\
            class AClass(object):
                def __init__(self, arg):
                    pass

                @staticmethod
                def create(*args, **kwds):
                    return AClass(*args, **kwds)
            a_var = AClass.create(10)
        """)
        self._introduce_factory(mod, mod.read().index("AClass") + 1, "create")
        self.assertEqual(expected, mod.read())

    def test_changing_occurrences_in_other_modules(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write("class AClass(object):\n    an_attr = 10\n")
        mod2.write("import mod1\na_var = mod1.AClass()\n")
        self._introduce_factory(mod1, mod1.read().index("AClass") + 1, "create")
        expected1 = dedent("""\
            class AClass(object):
                an_attr = 10

                @staticmethod
                def create(*args, **kwds):
                    return AClass(*args, **kwds)
        """)
        expected2 = dedent("""\
            import mod1
            a_var = mod1.AClass.create()
        """)
        self.assertEqual(expected1, mod1.read())
        self.assertEqual(expected2, mod2.read())

    def test_raising_exception_for_non_classes(self):
        mod = testutils.create_module(self.project, "mod")
        mod.write("def a_func():\n    pass\n")
        with self.assertRaises(RefactoringError):
            self._introduce_factory(mod, mod.read().index("a_func") + 1, "create")

    def test_undoing_introduce_factory(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        code1 = dedent("""\
            class AClass(object):
                an_attr = 10
        """)
        mod1.write(code1)
        code2 = dedent("""\
            from mod1 import AClass
            a_var = AClass()
        """)
        mod2.write(code2)
        self._introduce_factory(mod1, mod1.read().index("AClass") + 1, "create")
        self.project.history.undo()
        self.assertEqual(code1, mod1.read())
        self.assertEqual(code2, mod2.read())

    def test_using_on_an_occurrence_outside_the_main_module(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write("class AClass(object):\n    an_attr = 10\n")
        mod2.write("import mod1\na_var = mod1.AClass()\n")
        self._introduce_factory(mod2, mod2.read().index("AClass") + 1, "create")
        expected1 = dedent("""\
            class AClass(object):
                an_attr = 10

                @staticmethod
                def create(*args, **kwds):
                    return AClass(*args, **kwds)
        """)
        expected2 = "import mod1\n" "a_var = mod1.AClass.create()\n"
        self.assertEqual(expected1, mod1.read())
        self.assertEqual(expected2, mod2.read())

    def test_introduce_factory_in_nested_scopes(self):
        code = dedent("""\
            def create_var():
                class AClass(object):
                    an_attr = 10
                return AClass()
        """)
        mod = testutils.create_module(self.project, "mod")
        mod.write(code)
        expected = dedent("""\
            def create_var():
                class AClass(object):
                    an_attr = 10

                    @staticmethod
                    def create(*args, **kwds):
                        return AClass(*args, **kwds)
                return AClass.create()
        """)
        self._introduce_factory(mod, mod.read().index("AClass") + 1, "create")
        self.assertEqual(expected, mod.read())

    def test_adding_factory_for_global_factories(self):
        code = dedent("""\
            class AClass(object):
                an_attr = 10
        """)
        mod = testutils.create_module(self.project, "mod")
        mod.write(code)
        expected = dedent("""\
            class AClass(object):
                an_attr = 10

            def create(*args, **kwds):
                return AClass(*args, **kwds)
        """)
        self._introduce_factory(
            mod, mod.read().index("AClass") + 1, "create", global_factory=True
        )
        self.assertEqual(expected, mod.read())

    def test_get_name_for_factories(self):
        code = dedent("""\
            class C(object):
                pass
        """)
        mod = testutils.create_module(self.project, "mod")
        mod.write(code)
        factory = IntroduceFactory(self.project, mod, mod.read().index("C") + 1)
        self.assertEqual("C", factory.get_name())

    def test_raising_exception_for_global_factory_for_nested_classes(self):
        code = dedent("""\
            def create_var():
                class AClass(object):
                    an_attr = 10
                return AClass()
        """)
        mod = testutils.create_module(self.project, "mod")
        mod.write(code)
        with self.assertRaises(RefactoringError):
            self._introduce_factory(
                mod, mod.read().index("AClass") + 1, "create", global_factory=True
            )

    def test_changing_occurrences_in_the_main_module_for_global_factories(self):
        code = dedent("""\
            class AClass(object):
                an_attr = 10
            a_var = AClass()""")
        mod = testutils.create_module(self.project, "mod")
        mod.write(code)
        expected = dedent("""\
            class AClass(object):
                an_attr = 10

            def create(*args, **kwds):
                return AClass(*args, **kwds)
            a_var = create()""")
        self._introduce_factory(
            mod, mod.read().index("AClass") + 1, "create", global_factory=True
        )
        self.assertEqual(expected, mod.read())

    def test_changing_occurrences_in_other_modules_for_global_factories(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write("class AClass(object):\n    an_attr = 10\n")
        mod2.write("import mod1\na_var = mod1.AClass()\n")
        self._introduce_factory(
            mod1, mod1.read().index("AClass") + 1, "create", global_factory=True
        )
        expected1 = dedent("""\
            class AClass(object):
                an_attr = 10

            def create(*args, **kwds):
                return AClass(*args, **kwds)
        """)
        expected2 = "import mod1\n" "a_var = mod1.create()\n"
        self.assertEqual(expected1, mod1.read())
        self.assertEqual(expected2, mod2.read())

    def test_import_if_necessary_in_other_mods_for_global_factories(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod2 = testutils.create_module(self.project, "mod2")
        mod1.write("class AClass(object):\n    an_attr = 10\n")
        mod2.write("from mod1 import AClass\npair = AClass(), AClass\n")
        self._introduce_factory(
            mod1, mod1.read().index("AClass") + 1, "create", global_factory=True
        )
        expected1 = dedent("""\
            class AClass(object):
                an_attr = 10

            def create(*args, **kwds):
                return AClass(*args, **kwds)
        """)
        expected2 = dedent("""\
            from mod1 import AClass, create
            pair = create(), AClass
        """)
        self.assertEqual(expected1, mod1.read())
        self.assertEqual(expected2, mod2.read())

    def test_changing_occurrences_for_renamed_classes(self):
        code = dedent("""\
            class AClass(object):
                an_attr = 10
            a_class = AClass
            a_var = a_class()""")
        mod = testutils.create_module(self.project, "mod")
        mod.write(code)
        expected = dedent("""\
            class AClass(object):
                an_attr = 10

                @staticmethod
                def create(*args, **kwds):
                    return AClass(*args, **kwds)
            a_class = AClass
            a_var = a_class()""")
        self._introduce_factory(mod, mod.read().index("a_class") + 1, "create")
        self.assertEqual(expected, mod.read())

    def test_changing_occurs_in_the_same_module_with_conflict_ranges(self):
        mod = testutils.create_module(self.project, "mod")
        code = dedent("""\
            class C(object):
                def create(self):
                    return C()
        """)
        mod.write(code)
        self._introduce_factory(mod, mod.read().index("C"), "create_c", True)
        expected = dedent("""\
            class C(object):
                def create(self):
                    return create_c()
        """)
        self.assertTrue(mod.read().startswith(expected))

    def _transform_module_to_package(self, resource):
        self.project.do(
            rope.refactor.ModuleToPackage(self.project, resource).get_changes()
        )

    def test_transform_module_to_package(self):
        mod1 = testutils.create_module(self.project, "mod1")
        mod1.write("import mod2\nfrom mod2 import AClass\n")
        mod2 = testutils.create_module(self.project, "mod2")
        mod2.write("class AClass(object):\n    pass\n")
        self._transform_module_to_package(mod2)
        mod2 = self.project.get_resource("mod2")
        root_folder = self.project.root
        self.assertFalse(root_folder.has_child("mod2.py"))
        self.assertEqual(
            "class AClass(object):\n    pass\n",
            root_folder.get_child("mod2").get_child("__init__.py").read(),
        )

    def test_transform_module_to_package_undoing(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod = testutils.create_module(self.project, "mod", pkg)
        self._transform_module_to_package(mod)
        self.assertFalse(pkg.has_child("mod.py"))
        self.assertTrue(pkg.get_child("mod").has_child("__init__.py"))
        self.project.history.undo()
        self.assertTrue(pkg.has_child("mod.py"))
        self.assertFalse(pkg.has_child("mod"))

    def test_transform_module_to_package_with_relative_imports(self):
        pkg = testutils.create_package(self.project, "pkg")
        mod1 = testutils.create_module(self.project, "mod1", pkg)
        mod1.write("import mod2\nfrom mod2 import AClass\n")
        mod2 = testutils.create_module(self.project, "mod2", pkg)
        mod2.write("class AClass(object):\n    pass\n")
        self._transform_module_to_package(mod1)
        new_init = self.project.get_resource("pkg/mod1/__init__.py")
        self.assertEqual(
            "import pkg.mod2\nfrom pkg.mod2 import AClass\n", new_init.read()
        )

    def test_resources_parameter(self):
        code = dedent("""\
            class A(object):
                an_attr = 10
        """)
        code1 = dedent("""\
            import mod
            a = mod.A()
        """)
        mod = testutils.create_module(self.project, "mod")
        mod1 = testutils.create_module(self.project, "mod1")
        mod.write(code)
        mod1.write(code1)
        expected = dedent("""\
            class A(object):
                an_attr = 10

                @staticmethod
                def create(*args, **kwds):
                    return A(*args, **kwds)
        """)
        self._introduce_factory(
            mod, mod.read().index("A") + 1, "create", resources=[mod]
        )
        self.assertEqual(expected, mod.read())
        self.assertEqual(code1, mod1.read())


class EncapsulateFieldTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, "mod")
        self.mod1 = testutils.create_module(self.project, "mod1")
        self.a_class = dedent("""\
            class A(object):
                def __init__(self):
                    self.attr = 1
        """)
        self.added_methods = (
            "\n"
            "    def get_attr(self):\n"
            "        return self.attr\n\n"
            "    def set_attr(self, value):\n"
            "        self.attr = value\n"
        )
        self.encapsulated = self.a_class + self.added_methods

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _encapsulate(self, resource, offset, **args):
        changes = EncapsulateField(self.project, resource, offset).get_changes(**args)
        self.project.do(changes)

    def test_adding_getters_and_setters(self):
        code = self.a_class
        self.mod.write(code)
        self._encapsulate(self.mod, code.index("attr") + 1)
        self.assertEqual(self.encapsulated, self.mod.read())

    def test_changing_getters_in_other_modules(self):
        code = dedent("""\
            import mod
            a_var = mod.A()
            range(a_var.attr)
        """)
        self.mod1.write(code)
        self.mod.write(self.a_class)
        self._encapsulate(self.mod, self.mod.read().index("attr") + 1)
        expected = dedent("""\
            import mod
            a_var = mod.A()
            range(a_var.get_attr())
        """)
        self.assertEqual(expected, self.mod1.read())

    def test_changing_setters_in_other_modules(self):
        code = dedent("""\
            import mod
            a_var = mod.A()
            a_var.attr = 1
        """)
        self.mod1.write(code)
        self.mod.write(self.a_class)
        self._encapsulate(self.mod, self.mod.read().index("attr") + 1)
        expected = dedent("""\
            import mod
            a_var = mod.A()
            a_var.set_attr(1)
        """)
        self.assertEqual(expected, self.mod1.read())

    def test_changing_getters_in_setters(self):
        code = dedent("""\
            import mod
            a_var = mod.A()
            a_var.attr = 1 + a_var.attr
        """)
        self.mod1.write(code)
        self.mod.write(self.a_class)
        self._encapsulate(self.mod, self.mod.read().index("attr") + 1)
        expected = dedent("""\
            import mod
            a_var = mod.A()
            a_var.set_attr(1 + a_var.get_attr())
        """)
        self.assertEqual(expected, self.mod1.read())

    def test_appending_to_class_end(self):
        self.mod1.write(self.a_class + "a_var = A()\n")
        self._encapsulate(self.mod1, self.mod1.read().index("attr") + 1)
        self.assertEqual(self.encapsulated + "a_var = A()\n", self.mod1.read())

    def test_performing_in_other_modules(self):
        code = dedent("""\
            import mod
            a_var = mod.A()
            range(a_var.attr)
        """)
        self.mod1.write(code)
        self.mod.write(self.a_class)
        self._encapsulate(self.mod1, self.mod1.read().index("attr") + 1)
        self.assertEqual(self.encapsulated, self.mod.read())
        expected = dedent("""\
            import mod
            a_var = mod.A()
            range(a_var.get_attr())
        """)
        self.assertEqual(expected, self.mod1.read())

    def test_changing_main_module_occurrences(self):
        code = self.a_class + "a_var = A()\n" "a_var.attr = a_var.attr * 2\n"
        self.mod1.write(code)
        self._encapsulate(self.mod1, self.mod1.read().index("attr") + 1)
        expected = (
            self.encapsulated + "a_var = A()\n" "a_var.set_attr(a_var.get_attr() * 2)\n"
        )
        self.assertEqual(expected, self.mod1.read())

    def test_raising_exception_when_performed_on_non_attributes(self):
        self.mod1.write("attr = 10")
        with self.assertRaises(RefactoringError):
            self._encapsulate(self.mod1, self.mod1.read().index("attr") + 1)

    def test_raising_exception_on_tuple_assignments(self):
        self.mod.write(self.a_class)
        code = dedent("""\
            import mod
            a_var = mod.A()
            a_var.attr = 1
            a_var.attr, b = 1, 2
        """)
        self.mod1.write(code)
        with self.assertRaises(RefactoringError):
            self._encapsulate(self.mod1, self.mod1.read().index("attr") + 1)

    def test_raising_exception_on_tuple_assignments2(self):
        self.mod.write(self.a_class)
        code = dedent("""\
            import mod
            a_var = mod.A()
            a_var.attr = 1
            b, a_var.attr = 1, 2
        """)
        self.mod1.write(code)
        with self.assertRaises(RefactoringError):
            self._encapsulate(self.mod1, self.mod1.read().index("attr") + 1)

    def test_tuple_assignments_and_function_calls(self):
        code = dedent("""\
            import mod
            def func(a1=0, a2=0):
                pass
            a_var = mod.A()
            func(a_var.attr, a2=2)
        """)
        self.mod1.write(code)
        self.mod.write(self.a_class)
        self._encapsulate(self.mod, self.mod.read().index("attr") + 1)
        expected = dedent("""\
            import mod
            def func(a1=0, a2=0):
                pass
            a_var = mod.A()
            func(a_var.get_attr(), a2=2)
        """)
        self.assertEqual(expected, self.mod1.read())

    def test_tuple_assignments(self):
        code = dedent("""\
            import mod
            a_var = mod.A()
            a, b = a_var.attr, 1
        """)
        self.mod1.write(code)
        self.mod.write(self.a_class)
        self._encapsulate(self.mod, self.mod.read().index("attr") + 1)
        expected = dedent("""\
            import mod
            a_var = mod.A()
            a, b = a_var.get_attr(), 1
        """)
        self.assertEqual(expected, self.mod1.read())

    def test_changing_augmented_assignments(self):
        code = "import mod\n" "a_var = mod.A()\n" "a_var.attr += 1\n"
        self.mod1.write(code)
        self.mod.write(self.a_class)
        self._encapsulate(self.mod, self.mod.read().index("attr") + 1)
        expected = dedent("""\
            import mod
            a_var = mod.A()
            a_var.set_attr(a_var.get_attr() + 1)
        """)
        self.assertEqual(expected, self.mod1.read())

    def test_changing_augmented_assignments2(self):
        code = dedent("""\
            import mod
            a_var = mod.A()
            a_var.attr <<= 1
        """)
        self.mod1.write(code)
        self.mod.write(self.a_class)
        self._encapsulate(self.mod, self.mod.read().index("attr") + 1)
        expected = dedent("""\
            import mod
            a_var = mod.A()
            a_var.set_attr(a_var.get_attr() << 1)
        """)
        self.assertEqual(expected, self.mod1.read())

    def test_changing_occurrences_inside_the_class(self):
        new_class = (
            self.a_class + "\n" "    def a_func(self):\n" "        self.attr = 1\n"
        )
        self.mod.write(new_class)
        self._encapsulate(self.mod, self.mod.read().index("attr") + 1)
        expected = (
            self.a_class + "\n"
            "    def a_func(self):\n"
            "        self.set_attr(1)\n" + self.added_methods
        )
        self.assertEqual(expected, self.mod.read())

    def test_getter_and_setter_parameters(self):
        self.mod.write(self.a_class)
        self._encapsulate(
            self.mod,
            self.mod.read().index("attr") + 1,
            getter="getAttr",
            setter="setAttr",
        )
        new_methods = self.added_methods.replace("get_attr", "getAttr").replace(
            "set_attr", "setAttr"
        )
        expected = self.a_class + new_methods
        self.assertEqual(expected, self.mod.read())

    def test_using_resources_parameter(self):
        self.mod1.write("import mod\na = mod.A()\nvar = a.attr\n")
        self.mod.write(self.a_class)
        self._encapsulate(
            self.mod, self.mod.read().index("attr") + 1, resources=[self.mod]
        )
        self.assertEqual("import mod\na = mod.A()\nvar = a.attr\n", self.mod1.read())
        expected = self.a_class + self.added_methods
        self.assertEqual(expected, self.mod.read())


class LocalToFieldTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, "mod")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _perform_convert_local_variable_to_field(self, resource, offset):
        changes = LocalToField(self.project, resource, offset).get_changes()
        self.project.do(changes)

    def test_simple_local_to_field(self):
        code = dedent("""\
            class A(object):
                def a_func(self):
                    var = 10
        """)
        self.mod.write(code)
        self._perform_convert_local_variable_to_field(self.mod, code.index("var") + 1)
        expected = dedent("""\
            class A(object):
                def a_func(self):
                    self.var = 10
        """)
        self.assertEqual(expected, self.mod.read())

    def test_raising_exception_when_performed_on_a_global_var(self):
        self.mod.write("var = 10\n")
        with self.assertRaises(RefactoringError):
            self._perform_convert_local_variable_to_field(
                self.mod, self.mod.read().index("var") + 1
            )

    def test_raising_exception_when_performed_on_field(self):
        code = dedent("""\
            class A(object):
                def a_func(self):
                    self.var = 10
        """)
        self.mod.write(code)
        with self.assertRaises(RefactoringError):
            self._perform_convert_local_variable_to_field(
                self.mod, self.mod.read().index("var") + 1
            )

    def test_raising_exception_when_performed_on_a_parameter(self):
        code = dedent("""\
            class A(object):
                def a_func(self, var):
                    a = var
        """)
        self.mod.write(code)
        with self.assertRaises(RefactoringError):
            self._perform_convert_local_variable_to_field(
                self.mod, self.mod.read().index("var") + 1
            )

    # NOTE: This situation happens a lot and is normally not an error
    # @testutils.assert_raises(RefactoringError)
    def test_not_raise_exception_when_there_is_a_field_with_the_same_name(self):
        code = dedent("""\
            class A(object):
                def __init__(self):
                    self.var = 1
                def a_func(self):
                    var = 10
        """)
        self.mod.write(code)
        self._perform_convert_local_variable_to_field(
            self.mod, self.mod.read().rindex("var") + 1
        )

    def test_local_to_field_with_self_renamed(self):
        code = dedent("""\
            class A(object):
                def a_func(myself):
                    var = 10
        """)
        self.mod.write(code)
        self._perform_convert_local_variable_to_field(self.mod, code.index("var") + 1)
        expected = dedent("""\
            class A(object):
                def a_func(myself):
                    myself.var = 10
        """)
        self.assertEqual(expected, self.mod.read())


class IntroduceParameterTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore
        self.mod = testutils.create_module(self.project, "mod")

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def _introduce_parameter(self, offset, name):
        rope.refactor.introduce_parameter.IntroduceParameter(
            self.project, self.mod, offset
        ).get_changes(name).do()

    def test_simple_case(self):
        code = dedent("""\
            var = 1
            def f():
                b = var
        """)
        self.mod.write(code)
        offset = self.mod.read().rindex("var")
        self._introduce_parameter(offset, "var")
        expected = dedent("""\
            var = 1
            def f(var=var):
                b = var
        """)
        self.assertEqual(expected, self.mod.read())

    def test_changing_function_body(self):
        code = dedent("""\
            var = 1
            def f():
                b = var
        """)
        self.mod.write(code)
        offset = self.mod.read().rindex("var")
        self._introduce_parameter(offset, "p1")
        expected = dedent("""\
            var = 1
            def f(p1=var):
                b = p1
        """)
        self.assertEqual(expected, self.mod.read())

    def test_unknown_variables(self):
        self.mod.write("def f():\n    b = var + c\n")
        offset = self.mod.read().rindex("var")
        with self.assertRaises(RefactoringError):
            self._introduce_parameter(offset, "p1")
            self.assertEqual("def f(p1=var):\n    b = p1 + c\n", self.mod.read())

    def test_failing_when_not_inside(self):
        self.mod.write("var = 10\nb = var\n")
        offset = self.mod.read().rindex("var")
        with self.assertRaises(RefactoringError):
            self._introduce_parameter(offset, "p1")

    def test_attribute_accesses(self):
        code = dedent("""\
            class C(object):
                a = 10
            c = C()
            def f():
                b = c.a
        """)
        self.mod.write(code)
        offset = self.mod.read().rindex("a")
        self._introduce_parameter(offset, "p1")
        expected = dedent("""\
            class C(object):
                a = 10
            c = C()
            def f(p1=c.a):
                b = p1
        """)
        self.assertEqual(expected, self.mod.read())

    def test_introducing_parameters_for_methods(self):
        code = dedent("""\
            var = 1
            class C(object):
                def f(self):
                    b = var
        """)
        self.mod.write(code)
        offset = self.mod.read().rindex("var")
        self._introduce_parameter(offset, "p1")
        expected = dedent("""\
            var = 1
            class C(object):
                def f(self, p1=var):
                    b = p1
        """)
        self.assertEqual(expected, self.mod.read())


class _MockTaskObserver:
    def __init__(self):
        self.called = 0

    def __call__(self):
        self.called += 1


class TaskHandleTest(unittest.TestCase):
    def test_trivial_case(self):
        handle = rope.base.taskhandle.TaskHandle()
        self.assertFalse(handle.is_stopped())

    def test_stopping(self):
        handle = rope.base.taskhandle.TaskHandle()
        handle.stop()
        self.assertTrue(handle.is_stopped())

    def test_job_sets(self):
        handle = rope.base.taskhandle.TaskHandle()
        jobs = handle.create_jobset()
        self.assertEqual([jobs], handle.get_jobsets())

    def test_starting_and_finishing_jobs(self):
        handle = rope.base.taskhandle.TaskHandle()
        jobs = handle.create_jobset(name="test job set", count=1)
        jobs.started_job("job1")
        jobs.finished_job()

    def test_test_checking_status(self):
        handle = rope.base.taskhandle.TaskHandle()
        jobs = handle.create_jobset()
        handle.stop()
        with self.assertRaises(InterruptedTaskError):
            jobs.check_status()

    def test_test_checking_status_when_starting(self):
        handle = rope.base.taskhandle.TaskHandle()
        jobs = handle.create_jobset()
        handle.stop()
        with self.assertRaises(InterruptedTaskError):
            jobs.started_job("job1")

    def test_calling_the_observer_after_stopping(self):
        handle = rope.base.taskhandle.TaskHandle()
        observer = _MockTaskObserver()
        handle.add_observer(observer)
        handle.stop()
        self.assertEqual(1, observer.called)

    def test_calling_the_observer_after_creating_job_sets(self):
        handle = rope.base.taskhandle.TaskHandle()
        observer = _MockTaskObserver()
        handle.add_observer(observer)
        jobs = handle.create_jobset()  # noqa
        self.assertEqual(1, observer.called)

    def test_calling_the_observer_when_starting_and_finishing_jobs(self):
        handle = rope.base.taskhandle.TaskHandle()
        observer = _MockTaskObserver()
        handle.add_observer(observer)
        jobs = handle.create_jobset(name="test job set", count=1)
        jobs.started_job("job1")
        jobs.finished_job()
        self.assertEqual(3, observer.called)

    def test_job_set_get_percent_done(self):
        handle = rope.base.taskhandle.TaskHandle()
        jobs = handle.create_jobset(name="test job set", count=2)
        self.assertEqual(0, jobs.get_percent_done())
        jobs.started_job("job1")
        jobs.finished_job()
        self.assertEqual(50, jobs.get_percent_done())
        jobs.started_job("job2")
        jobs.finished_job()
        self.assertEqual(100, jobs.get_percent_done())

    def test_getting_job_name(self):
        handle = rope.base.taskhandle.TaskHandle()
        jobs = handle.create_jobset(name="test job set", count=1)
        # recommended name/job_name attribute
        self.assertEqual("test job set", jobs.name)
        self.assertEqual(None, jobs.job_name)
        # deprecated getters
        self.assertEqual("test job set", jobs.get_name())
        self.assertEqual(None, jobs.get_active_job_name())
        jobs.started_job("job1")
        # recommended name/job_name attribute
        self.assertEqual("test job set", jobs.get_name())
        # deprecated getters
        self.assertEqual("job1", jobs.get_active_job_name())



class ImportStatement:
    """Represent an import in a module

    `readonly` attribute controls whether this import can be changed
    by import actions or not.

    """

    def __init__(
        self, import_info, start_line, end_line, main_statement=None, blank_lines=0
    ):
        self.start_line = start_line
        self.end_line = end_line
        self.readonly = False
        self.main_statement = main_statement
        self._import_info = None
        self.import_info = import_info
        self._is_changed = False
        self.new_start = None
        self.blank_lines = blank_lines

    def _get_import_info(self):
        return self._import_info

    def _set_import_info(self, new_import):
        if (
            not self.readonly
            and new_import is not None
            and not new_import == self._import_info
        ):
            self._is_changed = True
            self._import_info = new_import

    import_info = property(_get_import_info, _set_import_info)

    def get_import_statement(self):
        if self._is_changed or self.main_statement is None:
            return self.import_info.get_import_statement()
        else:
            return self.main_statement

    def empty_import(self):
        self.import_info = ImportInfo.get_empty_import()

    def move(self, lineno, blank_lines=0):
        self.new_start = lineno
        self.blank_lines = blank_lines

    def get_old_location(self):
        return self.start_line, self.end_line

    def get_new_start(self):
        return self.new_start

    def is_changed(self):
        return self._is_changed or (
            self.new_start is not None or self.new_start != self.start_line
        )

    def accept(self, visitor):
        return visitor.dispatch(self)


class ImportInfo:
    def get_imported_primaries(self, context):
        pass

    def get_imported_names(self, context):
        return [
            primary.split(".")[0] for primary in self.get_imported_primaries(context)
        ]

    def get_import_statement(self):
        pass

    def is_empty(self):
        pass

    def __hash__(self):
        return hash(self.get_import_statement())

    def _are_name_and_alias_lists_equal(self, list1, list2):
        if len(list1) != len(list2):
            return False
        for pair1, pair2 in zip(list1, list2):
            if pair1 != pair2:
                return False
        return True

    def __eq__(self, obj):
        return (
            isinstance(obj, self.__class__)
            and self.get_import_statement() == obj.get_import_statement()
        )

    def __ne__(self, obj):
        return not self.__eq__(obj)

    @staticmethod
    def get_empty_import():
        return EmptyImport()


class NormalImport(ImportInfo):
    def __init__(self, names_and_aliases):
        self.names_and_aliases = names_and_aliases

    def get_imported_primaries(self, context):
        result = []
        for name, alias in self.names_and_aliases:
            if alias:
                result.append(alias)
            else:
                result.append(name)
        return result

    def get_import_statement(self):
        result = "import "
        for name, alias in self.names_and_aliases:
            result += name
            if alias:
                result += " as " + alias
            result += ", "
        return result[:-2]

    def is_empty(self):
        return len(self.names_and_aliases) == 0


class FromImport(ImportInfo):
    def __init__(self, module_name, level, names_and_aliases):
        self.module_name = module_name
        self.level = level
        self.names_and_aliases = names_and_aliases

    def get_imported_primaries(self, context):
        if self.names_and_aliases[0][0] == "*":
            module = self.get_imported_module(context)
            return [name for name in module if not name.startswith("_")]
        result = []
        for name, alias in self.names_and_aliases:
            if alias:
                result.append(alias)
            else:
                result.append(name)
        return result

    def get_imported_resource(self, context):
        """Get the imported resource

        Returns `None` if module was not found.
        """
        if self.level == 0:
            return context.project.find_module(self.module_name, folder=context.folder)
        else:
            return context.project.find_relative_module(
                self.module_name, context.folder, self.level
            )

    def get_imported_module(self, context):
        """Get the imported `PyModule`

        Raises `rope.base.exceptions.ModuleNotFoundError` if module
        could not be found.
        """
        if self.level == 0:
            return context.project.get_module(self.module_name, context.folder)
        else:
            return context.project.get_relative_module(
                self.module_name, context.folder, self.level
            )

    def get_import_statement(self):
        result = "from " + "." * self.level + self.module_name + " import "
        for name, alias in self.names_and_aliases:
            result += name
            if alias:
                result += " as " + alias
            result += ", "
        return result[:-2]

    def is_empty(self):
        return len(self.names_and_aliases) == 0

    def is_star_import(self):
        return len(self.names_and_aliases) > 0 and self.names_and_aliases[0][0] == "*"


class EmptyImport(ImportInfo):

    names_and_aliases = []

    def is_empty(self):
        return True

    def get_imported_primaries(self, context):
        return []


class ImportContext:
    def __init__(self, project, folder):
        self.project = project
        self.folder = folder



import secrets
from unittest.mock import patch

from rope.base import versioning


def test_calculate_version_hash(project):
    version_hash = versioning.calculate_version_hash(project)
    assert isinstance(version_hash, str)


def test_version_hash_is_constant(project):
    version_hash_1 = versioning.calculate_version_hash(project)
    version_hash_2 = versioning.calculate_version_hash(project)
    assert version_hash_1 == version_hash_2


def test_version_hash_varies_on_rope_version(project):
    actual_version_hash = versioning.calculate_version_hash(project)
    with patch("rope.VERSION", "1.0.0"):
        patched_version_hash = versioning.calculate_version_hash(project)
    assert actual_version_hash != patched_version_hash


def test_version_hash_varies_on_user_preferences(project):
    actual_version_hash = versioning.calculate_version_hash(project)
    assert project.prefs.get("automatic_soa") is False
    project.prefs.set("automatic_soa", True)
    patched_version_hash = versioning.calculate_version_hash(project)
    assert actual_version_hash != patched_version_hash


def test_version_hash_varies_on_get_file_content(project):
    actual_version_hash = versioning.calculate_version_hash(project)
    with patch("rope.base.versioning._get_file_content", return_value=secrets.token_hex()):
        patched_version_hash = versioning.calculate_version_hash(project)
    assert actual_version_hash != patched_version_hash



class ObjectDB:
    def __init__(self, db, validation):
        self.db = db
        self.validation = validation
        self.observers = []
        self.files = db.files

    def validate_files(self):
        for file in list(self.files):
            if not self.validation.is_file_valid(file):
                del self.files[file]
                self._file_removed(file)

    def validate_file(self, file):
        if file not in self.files:
            return
        for key in list(self.files[file]):
            if not self.validation.is_scope_valid(file, key):
                del self.files[file][key]

    def file_moved(self, file, newfile):
        if file not in self.files:
            return
        self.files.rename(file, newfile)
        self._file_removed(file)
        self._file_added(newfile)

    def get_files(self):
        return self.files.keys()

    def get_returned(self, path, key, args):
        scope_info = self._get_scope_info(path, key, readonly=True)
        result = scope_info.get_returned(args)
        if self.validation.is_value_valid(result):
            return result

    def get_pername(self, path, key, name):
        scope_info = self._get_scope_info(path, key, readonly=True)
        result = scope_info.get_per_name(name)
        if self.validation.is_value_valid(result):
            return result

    def get_callinfos(self, path, key):
        scope_info = self._get_scope_info(path, key, readonly=True)
        return scope_info.get_call_infos()

    def add_callinfo(self, path, key, args, returned):
        scope_info = self._get_scope_info(path, key, readonly=False)
        old_returned = scope_info.get_returned(args)
        if self.validation.is_more_valid(returned, old_returned):
            scope_info.add_call(args, returned)

    def add_pername(self, path, key, name, value):
        scope_info = self._get_scope_info(path, key, readonly=False)
        old_value = scope_info.get_per_name(name)
        if self.validation.is_more_valid(value, old_value):
            scope_info.save_per_name(name, value)

    def add_file_list_observer(self, observer):
        self.observers.append(observer)

    def write(self):
        self.db.write()

    def _get_scope_info(self, path, key, readonly=True):
        if path not in self.files:
            if readonly:
                return _NullScopeInfo()
            self.files.create(path)
            self._file_added(path)
        if key not in self.files[path]:
            if readonly:
                return _NullScopeInfo()
            self.files[path].create_scope(key)
        result = self.files[path][key]
        if isinstance(result, dict):
            print(self.files, self.files[path], self.files[path][key])
        return result

    def _file_removed(self, path):
        for observer in self.observers:
            observer.removed(path)

    def _file_added(self, path):
        for observer in self.observers:
            observer.added(path)

    def __str__(self):
        scope_count = 0
        for file_dict in self.files.values():
            scope_count += len(file_dict)
        return "ObjectDB holds {} file and {} scope infos".format(
            len(self.files),
            scope_count,
        )


class _NullScopeInfo:
    def __init__(self, error_on_write=True):
        self.error_on_write = error_on_write

    def get_per_name(self, name):
        pass

    def save_per_name(self, name, value):
        if self.error_on_write:
            raise NotImplementedError()

    def get_returned(self, parameters):
        pass

    def get_call_infos(self):
        return []

    def add_call(self, parameters, returned):
        if self.error_on_write:
            raise NotImplementedError()


class FileInfo(dict):
    def create_scope(self, key):
        pass


class FileDict(dict):
    def create(self, key):
        pass

    def rename(self, key, new_key):
        pass


class ScopeInfo:
    def get_per_name(self, name):
        pass

    def save_per_name(self, name, value):
        pass

    def get_returned(self, parameters):
        pass

    def get_call_infos(self):
        pass

    def add_call(self, parameters, returned):
        pass


class CallInfo:
    def __init__(self, args, returned):
        self.args = args
        self.returned = returned

    def get_parameters(self):
        return self.args

    def get_returned(self):
        return self.returned


class FileListObserver:
    def added(self, path):
        pass

    def removed(self, path):
        pass



from operator import itemgetter
from typing import Optional, Tuple

import rope.base.builtins
import rope.base.pynames
import rope.base.pyobjects
from rope.base import (
    arguments,
    ast,
    exceptions,
    nameanalyze,
    pyobjects,
    pyobjectsdef,
    worder,
)

BadIdentifierError = exceptions.BadIdentifierError


def eval_location(pymodule, offset):
    """Find the pyname at the offset"""
    return eval_location2(pymodule, offset)[1]


def eval_location2(pymodule, offset):
    """Find the primary and pyname at offset"""
    pyname_finder = ScopeNameFinder(pymodule)
    return pyname_finder.get_primary_and_pyname_at(offset)


def eval_node(scope, node):
    """Evaluate a `ast.AST` node and return a PyName

    Return `None` if the expression cannot be evaluated.
    """
    return eval_node2(scope, node)[1]


def eval_node2(scope, node):
    evaluator = StatementEvaluator(scope)
    evaluator.visit(node)
    return evaluator.old_result, evaluator.result


def eval_str(holding_scope, name):
    return eval_str2(holding_scope, name)[1]


def eval_str2(holding_scope, name):
    try:
        # parenthesizing for handling cases like 'a_var.\nattr'
        node = ast.parse("(%s)" % name)
    except SyntaxError:
        raise BadIdentifierError("Not a resolvable python identifier selected.")
    return eval_node2(holding_scope, node)


class ScopeNameFinder:
    def __init__(self, pymodule):
        self.module_scope = pymodule.get_scope()
        self.lines = pymodule.lines
        self.worder = worder.Worder(pymodule.source_code, True)

    def _is_defined_in_class_body(self, holding_scope, offset, lineno):
        if (
            lineno == holding_scope.get_start()
            and holding_scope.parent is not None
            and holding_scope.parent.get_kind() == "Class"
            and self.worder.is_a_class_or_function_name_in_header(offset)
        ):
            return True
        if (
            lineno != holding_scope.get_start()
            and holding_scope.get_kind() == "Class"
            and self.worder.is_name_assigned_in_class_body(offset)
        ):
            return True
        return False

    def _is_function_name_in_function_header(self, scope, offset, lineno):
        return (
            scope.get_start() <= lineno <= scope.get_body_start()
            and scope.get_kind() == "Function"
            and self.worder.is_a_class_or_function_name_in_header(offset)
        )

    def get_pyname_at(self, offset):
        return self.get_primary_and_pyname_at(offset)[1]

    def get_primary_and_pyname_at(
        self,
        offset: int,
    ) -> Tuple[Optional[rope.base.pynames.PyName], Optional[rope.base.pynames.PyName]]:
        lineno = self.lines.get_line_number(offset)
        holding_scope = self.module_scope.get_inner_scope_for_offset(offset)
        # function keyword parameter
        if self.worder.is_function_keyword_parameter(offset):
            keyword_name = self.worder.get_word_at(offset)
            pyobject = self.get_enclosing_function(offset)
            if isinstance(pyobject, pyobjectsdef.PyFunction):
                parameter_name = pyobject.get_parameters().get(keyword_name, None)
                return (None, parameter_name)
            elif isinstance(pyobject, pyobjects.AbstractFunction):
                parameter_name = rope.base.pynames.ParameterName()
                return (None, parameter_name)
        # class body
        if self._is_defined_in_class_body(holding_scope, offset, lineno):
            class_scope = holding_scope
            if lineno == holding_scope.get_start():
                class_scope = holding_scope.parent
            name = self.worder.get_primary_at(offset).strip()
            try:
                return (None, class_scope.pyobject[name])
            except rope.base.exceptions.AttributeNotFoundError:
                return (None, None)
        # function header
        if self._is_function_name_in_function_header(holding_scope, offset, lineno):
            name = self.worder.get_primary_at(offset).strip()
            return (None, holding_scope.parent[name])
        # module in a from statement or an imported name that is aliased
        if self.worder.is_from_statement_module(
            offset
        ) or self.worder.is_import_statement_aliased_module(offset):
            module = self.worder.get_primary_at(offset)
            module_pyname = self._find_module(module)
            return (None, module_pyname)
        if self.worder.is_from_aliased(offset):
            name = self.worder.get_from_aliased(offset)
        else:
            name = self.worder.get_primary_at(offset)
        return eval_str2(holding_scope, name)

    def get_enclosing_function(self, offset):
        function_parens = self.worder.find_parens_start_from_inside(offset)
        try:
            function_pyname = self.get_pyname_at(function_parens - 1)
        except BadIdentifierError:
            function_pyname = None
        if function_pyname is not None:
            pyobject = function_pyname.get_object()
            if isinstance(pyobject, pyobjects.AbstractFunction):
                return pyobject
            elif (
                isinstance(pyobject, pyobjects.AbstractClass) and "__init__" in pyobject
            ):
                return pyobject["__init__"].get_object()
            elif "__call__" in pyobject:
                return pyobject["__call__"].get_object()
        return None

    def _find_module(self, module_name):
        dots = 0
        while module_name[dots] == ".":
            dots += 1
        return rope.base.pynames.ImportedModule(
            self.module_scope.pyobject, module_name[dots:], dots
        )


class StatementEvaluator(ast.RopeNodeVisitor):
    def __init__(self, scope):
        self.scope = scope
        self.result = None
        self.old_result = None

    def _Name(self, node):
        self.result = self.scope.lookup(node.id)

    def _Attribute(self, node):
        pyname = eval_node(self.scope, node.value)
        if pyname is None:
            pyname = rope.base.pynames.UnboundName()
        self.old_result = pyname
        if pyname.get_object() != rope.base.pyobjects.get_unknown():
            try:
                self.result = pyname.get_object()[node.attr]
            except exceptions.AttributeNotFoundError:
                self.result = None

    def _Call(self, node):
        primary, pyobject = self._get_primary_and_object_for_node(node.func)
        if pyobject is None:
            return

        def _get_returned(pyobject):
            args = arguments.create_arguments(primary, pyobject, node, self.scope)
            return pyobject.get_returned_object(args)

        if isinstance(pyobject, rope.base.pyobjects.AbstractClass):
            result = None
            if "__new__" in pyobject:
                new_function = pyobject["__new__"].get_object()
                result = _get_returned(new_function)
            if result is None or result == rope.base.pyobjects.get_unknown():
                result = rope.base.pyobjects.PyObject(pyobject)
            self.result = rope.base.pynames.UnboundName(pyobject=result)
            return

        pyfunction = None
        if isinstance(pyobject, rope.base.pyobjects.AbstractFunction):
            pyfunction = pyobject
        elif "__call__" in pyobject:
            pyfunction = pyobject["__call__"].get_object()
        if pyfunction is not None:
            self.result = rope.base.pynames.UnboundName(
                pyobject=_get_returned(pyfunction)
            )

    def _Str(self, node):
        self.result = rope.base.pynames.UnboundName(
            pyobject=rope.base.builtins.get_str()
        )

    def _Num(self, node):
        type_name = type(node.n).__name__
        self.result = self._get_builtin_name(type_name)

    def _Constant(self, node):
        type_name = type(node.n).__name__
        try:
            self.result = self._get_builtin_name(type_name)
        except exceptions.AttributeNotFoundError:
            # XXX: Right way to fix this is to add missing NoneType to builtins?
            pass

    def _get_builtin_name(self, type_name):
        pytype = rope.base.builtins.builtins[type_name].get_object()
        return rope.base.pynames.UnboundName(rope.base.pyobjects.PyObject(pytype))

    def _BinOp(self, node):
        self.result = rope.base.pynames.UnboundName(
            self._get_object_for_node(node.left)
        )

    def _BoolOp(self, node):
        pyobject = self._get_object_for_node(node.values[0])
        if pyobject is None:
            pyobject = self._get_object_for_node(node.values[1])
        self.result = rope.base.pynames.UnboundName(pyobject)

    def _Repr(self, node):
        self.result = self._get_builtin_name("str")

    def _UnaryOp(self, node):
        self.result = rope.base.pynames.UnboundName(
            self._get_object_for_node(node.operand)
        )

    def _Compare(self, node):
        self.result = self._get_builtin_name("bool")

    def _Dict(self, node):
        keys = None
        values = None
        if node.keys and node.keys[0]:
            keys, values = next(
                iter(filter(itemgetter(0), zip(node.keys, node.values))), (None, None)
            )
            if keys:
                keys = self._get_object_for_node(keys)
            if values:
                values = self._get_object_for_node(values)
        self.result = rope.base.pynames.UnboundName(
            pyobject=rope.base.builtins.get_dict(keys, values)
        )

    def _List(self, node):
        holding = None
        if node.elts:
            holding = self._get_object_for_node(node.elts[0])
        self.result = rope.base.pynames.UnboundName(
            pyobject=rope.base.builtins.get_list(holding)
        )

    def _ListComp(self, node):
        pyobject = self._what_does_comprehension_hold(node)
        self.result = rope.base.pynames.UnboundName(
            pyobject=rope.base.builtins.get_list(pyobject)
        )

    def _GeneratorExp(self, node):
        pyobject = self._what_does_comprehension_hold(node)
        self.result = rope.base.pynames.UnboundName(
            pyobject=rope.base.builtins.get_iterator(pyobject)
        )

    def _what_does_comprehension_hold(self, node):
        scope = self._make_comprehension_scope(node)
        pyname = eval_node(scope, node.elt)
        return pyname.get_object() if pyname is not None else None

    def _make_comprehension_scope(self, node):
        scope = self.scope
        module = scope.pyobject.get_module()
        names = {}
        for comp in node.generators:
            new_names = _get_evaluated_names(
                comp.target, comp.iter, module, ".__iter__().next()", node.lineno
            )
            names.update(new_names)
        return rope.base.pyscopes.TemporaryScope(scope.pycore, scope, names)

    def _Tuple(self, node):
        objects = []
        if len(node.elts) < 4:
            for stmt in node.elts:
                pyobject = self._get_object_for_node(stmt)
                objects.append(pyobject)
        else:
            objects.append(self._get_object_for_node(node.elts[0]))
        self.result = rope.base.pynames.UnboundName(
            pyobject=rope.base.builtins.get_tuple(*objects)
        )

    def _get_object_for_node(self, stmt):
        pyname = eval_node(self.scope, stmt)
        pyobject = None
        if pyname is not None:
            pyobject = pyname.get_object()
        return pyobject

    def _get_primary_and_object_for_node(self, stmt):
        primary, pyname = eval_node2(self.scope, stmt)
        pyobject = None
        if pyname is not None:
            pyobject = pyname.get_object()
        return primary, pyobject

    def _Subscript(self, node):
        if isinstance(node.slice, ast.Index):
            self._call_function(node.value, "__getitem__", [node.slice.value])
        elif isinstance(node.slice, ast.Slice):
            self._call_function(node.value, "__getitem__", [node.slice])
        elif isinstance(node.slice, ast.expr):
            self._call_function(node.value, "__getitem__", [node.value])

    def _Slice(self, node):
        self.result = self._get_builtin_name("slice")

    def _call_function(self, node, function_name, other_args=None):
        pyname = eval_node(self.scope, node)
        if pyname is not None:
            pyobject = pyname.get_object()
        else:
            return
        if function_name in pyobject:
            called = pyobject[function_name].get_object()
            if not called or not isinstance(called, pyobjects.AbstractFunction):
                return
            args = [node]
            if other_args:
                args += other_args
            arguments_ = arguments.Arguments(args, self.scope)
            self.result = rope.base.pynames.UnboundName(
                pyobject=called.get_returned_object(arguments_)
            )

    def _Lambda(self, node):
        self.result = rope.base.pynames.UnboundName(
            pyobject=rope.base.builtins.Lambda(node, self.scope)
        )


def _get_evaluated_names(targets, assigned, module, evaluation, lineno):
    result = {}
    for name, levels in nameanalyze.get_name_levels(targets):
        assignment = rope.base.pynames.AssignmentValue(assigned, levels, evaluation)
        # XXX: this module should not access `rope.base.pynamesdef`!
        pyname = rope.base.pynamesdef.AssignedName(lineno, module)
        pyname.assignments.append(assignment)
        result[name] = pyname
    return result
